from tests.bofire.data_models.specs.acquisition_functions import (
    specs as acquisition_functions,
)
from tests.bofire.data_models.specs.constraints import specs as constraints
from tests.bofire.data_models.specs.domain import specs as domain
from tests.bofire.data_models.specs.features import specs as features
from tests.bofire.data_models.specs.objectives import specs as objectives
from tests.bofire.data_models.specs.specs import Spec, Specs

try:
    # in case of the minimal installation these import are not available
    from tests.bofire.data_models.specs.conditions import specs as conditions
    from tests.bofire.data_models.specs.kernels import specs as kernels
    from tests.bofire.data_models.specs.molfeatures import (
        specs as molfeatures,
    )
    from tests.bofire.data_models.specs.outlier_detection import (
        specs as outlier_detection,
    )
    from tests.bofire.data_models.specs.priors import specs as priors
    from tests.bofire.data_models.specs.strategies import specs as strategies
    from tests.bofire.data_models.specs.surrogates import specs as surrogates
except ImportError:
    pass



from pydantic import validator

from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MinimizeObjective,
)
from bofire.data_models.strategies.predictives.botorch import BotorchStrategy


class MultiobjectiveStrategy(BotorchStrategy):
    @validator("domain")
    def validate_domain_is_multiobjective(cls, v, values):
        """Validate that the domain is multiobjective."""
        feats = v.outputs.get_by_objective(
            [MaximizeObjective, MinimizeObjective, CloseToTargetObjective]
        )
        if len(feats) < 2:
            raise ValueError(
                "At least two output features with MaximizeObjective or MinimizeObjective has to be defined in the domain."
            )
        for feat in feats:
            if feat.objective.w != 1.0:  # type: ignore
                raise ValueError(
                    f"Only objectives with weight 1 are supported. Violated by feature {feat.key}."
                )
        return v



from typing import List, Optional

import numpy as np
import torch
from botorch.acquisition import AcquisitionFunction, get_acquisition_function
from botorch.acquisition.multi_objective.objective import (
    GenericMCMultiOutputObjective,
    MCMultiOutputObjective,
)
from botorch.models.gpytorch import GPyTorchModel

from bofire.data_models.acquisition_functions.api import (
    qEHVI,
    qLogEHVI,
    qLogNEHVI,
    qNEHVI,
)
from bofire.data_models.objectives.api import ConstrainedObjective
from bofire.data_models.strategies.api import MoboStrategy as DataModel
from bofire.strategies.predictives.botorch import BotorchStrategy
from bofire.utils.multiobjective import get_ref_point_mask, infer_ref_point
from bofire.utils.torch_tools import (
    get_multiobjective_objective,
    get_output_constraints,
    tkwargs,
)


class MoboStrategy(BotorchStrategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.ref_point = data_model.ref_point
        self.ref_point_mask = get_ref_point_mask(self.domain)
        self.acquisition_function = data_model.acquisition_function

    ref_point: Optional[dict] = None
    objective: Optional[MCMultiOutputObjective] = None

    def _get_acqfs(self, n) -> List[AcquisitionFunction]:
        assert self.is_fitted is True, "Model not trained."

        X_train, X_pending = self.get_acqf_input_tensors()

        # get etas and constraints
        constraints, etas = get_output_constraints(self.domain.outputs)
        if len(constraints) == 0:
            constraints, etas = None, 1e-3
        else:
            etas = torch.tensor(etas).to(**tkwargs)

        objective = self._get_objective()
        # in case that qehvi, qlogehvi is used we need also y
        if isinstance(self.acquisition_function, (qLogEHVI, qEHVI)):
            Y = torch.from_numpy(
                self.domain.outputs.preprocess_experiments_all_valid_outputs(
                    self.experiments
                )[self.domain.outputs.get_keys()].values
            ).to(**tkwargs)
        else:
            Y = None

        assert self.model is not None

        acqf = get_acquisition_function(
            self.acquisition_function.__class__.__name__,
            self.model,
            ref_point=self.get_adjusted_refpoint(),
            objective=objective,
            X_observed=X_train,
            X_pending=X_pending,
            constraints=constraints,
            eta=etas,
            mc_samples=self.num_sobol_samples,
            cache_root=True if isinstance(self.model, GPyTorchModel) else False,
            alpha=self.acquisition_function.alpha,
            prune_baseline=self.acquisition_function.prune_baseline
            if isinstance(self.acquisition_function, (qLogNEHVI, qNEHVI))
            else True,
            Y=Y,
        )
        return [acqf]

    # def _get_acqfs(
    #     self, n
    # ) -> List[
    #     Union[
    #         qExpectedHypervolumeImprovement,
    #         qNoisyExpectedHypervolumeImprovement,
    #         qLogNoisyExpectedHypervolumeImprovement,
    #         qLogExpectedHypervolumeImprovement,
    #     ]
    # ]:
    #     df = self.domain.outputs.preprocess_experiments_all_valid_outputs(
    #         self.experiments
    #     )

    #     train_obj = (
    #         df[self.domain.outputs.get_keys_by_objective(excludes=None)].values
    #         * self.ref_point_mask
    #     )
    #     ref_point = self.get_adjusted_refpoint()
    #     weights = np.array(
    #         [
    #             feat.objective.w  # type: ignore
    #             for feat in self.domain.outputs.get_by_objective(excludes=None)
    #         ]
    #     )
    #     # compute points that are better than the known reference point
    #     better_than_ref = (train_obj > ref_point).all(axis=-1)
    #     # partition non-dominated space into disjoint rectangles
    #     partitioning = NondominatedPartitioning(
    #         ref_point=torch.from_numpy(ref_point * weights),
    #         # use observations that are better than the specified reference point and feasible
    #         Y=torch.from_numpy(train_obj[better_than_ref]),
    #     )

    #     _, X_pending = self.get_acqf_input_tensors()

    #     assert self.model is not None
    #     # setup the acqf
    #     acqf = qExpectedHypervolumeImprovement(
    #         model=self.model,
    #         ref_point=ref_point,  # use known reference point
    #         partitioning=partitioning,
    #         # sampler=self.sampler,
    #         # define an objective that specifies which outcomes are the objectives
    #         objective=self._get_objective(),
    #         X_pending=X_pending,
    #     )
    #     acqf._default_sample_shape = torch.Size([self.num_sobol_samples])
    #     return [acqf]

    def _get_objective(self) -> GenericMCMultiOutputObjective:
        objective = get_multiobjective_objective(outputs=self.domain.outputs)
        return GenericMCMultiOutputObjective(objective=objective)

    def get_adjusted_refpoint(self) -> List[float]:
        if self.ref_point is None:
            df = self.domain.outputs.preprocess_experiments_all_valid_outputs(
                self.experiments
            )
            ref_point = infer_ref_point(
                self.domain, experiments=df, return_masked=False
            )
        else:
            ref_point = self.ref_point
        return (
            self.ref_point_mask
            * np.array(
                [
                    ref_point[feat]
                    for feat in self.domain.outputs.get_keys_by_objective(
                        excludes=ConstrainedObjective
                    )
                ]
            )
        ).tolist()



from typing import Literal

import numpy as np
import pandas as pd
import pytest
from pydantic.error_wrappers import ValidationError

from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.surrogates.api import Surrogate as SurrogateDataModel
from bofire.surrogates.api import PredictedValue, Surrogate


class DummyDataModel(SurrogateDataModel):
    type: Literal["Dummy"] = "Dummy"


class Dummy(Surrogate):
    def __init__(
        self,
        data_model: SurrogateDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _predict(self, transformed_X: pd.DataFrame):
        preds = np.random.normal(
            loc=5, scale=1, size=(len(transformed_X), len(self.outputs))
        )
        stds = np.random.uniform(
            low=0.0, high=1.0, size=(len(transformed_X), len(self.outputs))
        )
        return preds, stds

    def _dumps(self):
        pass

    def loads(self, dumpstr: str):
        pass


def test_zero_inputs():
    inputs = Inputs(features=[])
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    with pytest.raises(ValueError):
        data_model = DummyDataModel(inputs=inputs, outputs=outputs)
        Dummy(data_model=data_model)


def test_zero_outputs():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(5)
        ]
    )
    outputs = Outputs(features=[])
    with pytest.raises(ValueError):
        data_model = DummyDataModel(inputs=inputs, outputs=outputs)
        Dummy(data_model=data_model)


def test_predicted_value():
    # valid
    PredictedValue(predictedValue=5.0, standardDeviation=0.1)
    PredictedValue(predictedValue=4.2, standardDeviation=0.01)
    # invalid
    with pytest.raises(ValidationError):
        PredictedValue(predictedValue=5.0, standardDeviation=-0.1)


@pytest.mark.parametrize("n_outputs", [1, 2])
def test_to_outputs(n_outputs):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(5)
        ]
    )
    outputs = Outputs(
        features=[ContinuousOutput(key=f"y_{i+1}") for i in range(n_outputs)]
    )
    data_model = DummyDataModel(inputs=inputs, outputs=outputs)
    model = Dummy(data_model=data_model)

    model.model = "dummymodel"  # type: ignore
    preds = model.predict(inputs.sample(10))
    output = model.to_predictions(preds)
    assert len(output) == n_outputs
    assert sorted(output.keys()) == [f"y_{i+1}" for i in range(n_outputs)]
    for key in [f"y_{i+1}" for i in range(n_outputs)]:
        assert len(output[key]) == 10


def test_is_fitted():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(5)
        ]
    )

    outputs = Outputs(features=[ContinuousOutput(key="y")])
    data_model = DummyDataModel(inputs=inputs, outputs=outputs)
    d = Dummy(data_model=data_model)
    assert d.is_fitted is False
    d.model = "dummymodel"  # type: ignore
    assert d.is_fitted is True



from typing import Dict, Optional

import numpy as np
import pandas as pd
import torch
from botorch.utils.multi_objective import is_non_dominated
from botorch.utils.multi_objective.hypervolume import Hypervolume

from bofire.data_models.domain.api import Domain
from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MinimizeObjective,
)
from bofire.utils.torch_tools import get_multiobjective_objective, tkwargs


def get_ref_point_mask(
    domain: Domain, output_feature_keys: Optional[list] = None
) -> np.ndarray:
    """Method to get a mask for the reference points taking into account if we
    want to maximize or minimize an objective. In case it is maximize the value
    in the mask is 1, in case we want to minimize it is -1.

    Args:
        domain (Domain): Domain for which the mask should be generated.
        output_feature_keys (Optional[list], optional): Name of output feature keys
            that should be considered in the mask. Defaults to None.

    Returns:
        np.ndarray: _description_
    """
    if output_feature_keys is None:
        output_feature_keys = domain.outputs.get_keys_by_objective(
            includes=[MaximizeObjective, MinimizeObjective, CloseToTargetObjective]
        )
    if len(output_feature_keys) < 2:
        raise ValueError("At least two output features have to be provided.")
    mask = []
    for key in output_feature_keys:
        feat = domain.get_feature(key)
        if isinstance(feat.objective, MaximizeObjective):  # type: ignore
            mask.append(1.0)
        elif isinstance(feat.objective, MinimizeObjective):  # type: ignore
            mask.append(-1.0)
        elif isinstance(feat.objective, CloseToTargetObjective):  # type: ignore
            mask.append(-1.0)
        else:
            raise ValueError(
                "Only `MaximizeObjective` and `MinimizeObjective` supported"
            )
    return np.array(mask)


def get_pareto_front(
    domain: Domain,
    experiments: pd.DataFrame,
    output_feature_keys: Optional[list] = None,
) -> pd.DataFrame:
    if output_feature_keys is None:
        outputs = domain.outputs.get_by_objective(
            includes=[MaximizeObjective, MinimizeObjective, CloseToTargetObjective]
        )
    else:
        outputs = domain.outputs.get_by_keys(output_feature_keys)
    assert len(outputs) >= 2, "At least two output features have to be provided."
    output_feature_keys = [f.key for f in outputs]
    df = domain.outputs.preprocess_experiments_all_valid_outputs(
        experiments, output_feature_keys
    )
    objective = get_multiobjective_objective(outputs=outputs)  # type: ignore
    pareto_mask = np.array(
        is_non_dominated(
            objective(
                torch.from_numpy(df[output_feature_keys].values).to(**tkwargs), None
            )
        )
    )
    return df.loc[pareto_mask]


def compute_hypervolume(
    domain: Domain, optimal_experiments: pd.DataFrame, ref_point: dict
) -> float:
    outputs = domain.outputs.get_by_objective(
        includes=[MaximizeObjective, MinimizeObjective, CloseToTargetObjective]
    )
    objective = get_multiobjective_objective(outputs=outputs)
    ref_point_mask = torch.from_numpy(get_ref_point_mask(domain)).to(**tkwargs)
    hv = Hypervolume(
        ref_point=torch.tensor(
            [
                ref_point[feat]
                for feat in domain.outputs.get_keys_by_objective(
                    includes=[
                        MaximizeObjective,
                        MinimizeObjective,
                        CloseToTargetObjective,
                    ]
                )
            ]
        ).to(**tkwargs)
        * ref_point_mask
    )

    return hv.compute(
        objective(  # type: ignore
            torch.from_numpy(
                optimal_experiments[
                    domain.outputs.get_keys_by_objective(
                        includes=[
                            MaximizeObjective,
                            MinimizeObjective,
                            CloseToTargetObjective,
                        ]
                    )
                ].values
            ).to(**tkwargs)
        )
    )


def infer_ref_point(
    domain: Domain, experiments: pd.DataFrame, return_masked: bool = False
) -> Dict[str, float]:
    outputs = domain.outputs.get_by_objective(
        includes=[MaximizeObjective, MinimizeObjective, CloseToTargetObjective]
    )
    keys = [f.key for f in outputs]
    objective = get_multiobjective_objective(outputs=outputs)

    df = domain.outputs.preprocess_experiments_all_valid_outputs(
        experiments, output_feature_keys=keys
    )

    ref_point_array = (
        objective(torch.from_numpy(df[keys].values).to(**tkwargs), None)
        .numpy()
        .min(axis=0)
    )

    mask = get_ref_point_mask(domain)
    # ref_point_array = (df[keys].values * mask).min(axis=0)
    if return_masked is False:
        ref_point_array /= mask
    return {feat: ref_point_array[i] for i, feat in enumerate(keys)}



from typing import Literal, Optional

import pandas as pd
from pydantic import Field

from bofire.data_models.domain.api import Inputs
from bofire.data_models.enum import RegressionMetricsEnum
from bofire.data_models.features.api import CategoricalInput
from bofire.data_models.kernels.api import (
    AnyKernel,
    MaternKernel,
    RBFKernel,
    ScaleKernel,
)
from bofire.data_models.priors.api import (
    BOTORCH_LENGTHCALE_PRIOR,
    BOTORCH_NOISE_PRIOR,
    BOTORCH_SCALE_PRIOR,
    MBO_LENGTHCALE_PRIOR,
    MBO_NOISE_PRIOR,
    MBO_OUTPUTSCALE_PRIOR,
    AnyPrior,
)

# from bofire.data_models.strategies.api import FactorialStrategy
from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import Hyperconfig, TrainableSurrogate


class SingleTaskGPHyperconfig(Hyperconfig):
    type: Literal["SingleTaskGPHyperconfig"] = "SingleTaskGPHyperconfig"
    inputs: Inputs = Inputs(
        features=[
            CategoricalInput(
                key="kernel", categories=["rbf", "matern_1.5", "matern_2.5"]
            ),
            CategoricalInput(key="prior", categories=["mbo", "botorch"]),
            CategoricalInput(key="ard", categories=["True", "False"]),
        ]
    )
    target_metric = RegressionMetricsEnum.MAE
    hyperstrategy: Literal[
        "FactorialStrategy", "SoboStrategy", "RandomStrategy"
    ] = "FactorialStrategy"

    @staticmethod
    def _update_hyperparameters(
        surrogate_data: "SingleTaskGPSurrogate", hyperparameters: pd.Series
    ):
        def matern_25(ard: bool, lengthscale_prior: AnyPrior) -> MaternKernel:
            return MaternKernel(nu=2.5, lengthscale_prior=lengthscale_prior, ard=ard)

        def matern_15(ard: bool, lengthscale_prior: AnyPrior) -> MaternKernel:
            return MaternKernel(nu=1.5, lengthscale_prior=lengthscale_prior, ard=ard)

        if hyperparameters.prior == "mbo":
            noise_prior, lengthscale_prior, outputscale_prior = (
                MBO_NOISE_PRIOR(),
                MBO_LENGTHCALE_PRIOR(),
                MBO_OUTPUTSCALE_PRIOR(),
            )
        else:
            noise_prior, lengthscale_prior, outputscale_prior = (
                BOTORCH_NOISE_PRIOR(),
                BOTORCH_LENGTHCALE_PRIOR(),
                BOTORCH_SCALE_PRIOR(),
            )
        surrogate_data.noise_prior = noise_prior
        if hyperparameters.kernel == "rbf":
            surrogate_data.kernel = ScaleKernel(
                base_kernel=RBFKernel(
                    ard=hyperparameters.ard, lengthscale_prior=lengthscale_prior
                ),
                outputscale_prior=outputscale_prior,
            )
        elif hyperparameters.kernel == "matern_2.5":
            surrogate_data.kernel = ScaleKernel(
                base_kernel=matern_25(
                    ard=hyperparameters.ard, lengthscale_prior=lengthscale_prior
                ),
                outputscale_prior=outputscale_prior,
            )
        elif hyperparameters.kernel == "matern_1.5":
            surrogate_data.kernel = ScaleKernel(
                base_kernel=matern_15(
                    ard=hyperparameters.ard, lengthscale_prior=lengthscale_prior
                ),
                outputscale_prior=outputscale_prior,
            )
        else:
            raise ValueError(f"Kernel {hyperparameters.kernel} not known.")


class SingleTaskGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["SingleTaskGPSurrogate"] = "SingleTaskGPSurrogate"

    kernel: AnyKernel = Field(
        default_factory=lambda: ScaleKernel(
            base_kernel=MaternKernel(
                ard=True,
                nu=2.5,
                lengthscale_prior=BOTORCH_LENGTHCALE_PRIOR(),
            ),
            outputscale_prior=BOTORCH_SCALE_PRIOR(),
        )
    )
    noise_prior: AnyPrior = Field(default_factory=lambda: BOTORCH_NOISE_PRIOR())
    scaler: ScalerEnum = ScalerEnum.NORMALIZE
    hyperconfig: Optional[SingleTaskGPHyperconfig] = Field(
        default_factory=lambda: SingleTaskGPHyperconfig()
    )



import math
from itertools import combinations
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from bofire.data_models.constraints.linear import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
)
from bofire.data_models.constraints.nchoosek import NChooseKConstraint
from bofire.data_models.constraints.nonlinear import NonlinearInequalityConstraint
from bofire.data_models.domain.domain import Domain
from bofire.data_models.features.categorical import CategoricalInput
from bofire.data_models.features.continuous import ContinuousInput
from bofire.data_models.features.discrete import DiscreteInput
from bofire.data_models.features.feature import Feature, Output, TDiscreteVals


def discrete_to_relaxable_domain_mapper(
    domain: Domain,
) -> Tuple[
    Domain,
    List[List[ContinuousInput]],
    Dict[str, Tuple[ContinuousInput, TDiscreteVals]],
]:
    """Converts a domain with discrete and categorical inputs to a domain with relaxable inputs.

    Args:
        domain (Domain): Domain with discrete and categorical inputs.
    """

    # get all discrete and categorical inputs
    kept_inputs = domain.get_features(
        excludes=[CategoricalInput, DiscreteInput, Output]
    ).features
    discrete_inputs = domain.inputs.get(DiscreteInput)
    categorical_inputs = domain.inputs.get(CategoricalInput)

    # convert discrete inputs to continuous inputs
    relaxable_discrete_inputs = {
        d_input.key: (  # type: ignore
            ContinuousInput(key=d_input.key, bounds=(min(d_input.values), max(d_input.values))),  # type: ignore
            d_input.values,  # type: ignore
        )  # type: ignore
        for d_input in discrete_inputs
    }

    # convert categorical inputs to continuous inputs
    relaxable_categorical_inputs = []
    new_constraints = []
    categorical_groups: List[List[ContinuousInput]] = []
    for c_input in categorical_inputs:
        current_group_keys = list(c_input.categories)  # type: ignore
        pick_1_constraint, group_vars = generate_mixture_constraints(current_group_keys)
        categorical_groups.append(group_vars)
        relaxable_categorical_inputs.extend(group_vars)
        new_constraints.append(pick_1_constraint)

    # create new domain with continuous inputs
    new_domain = Domain(
        inputs=kept_inputs + [var for key, (var, values) in relaxable_discrete_inputs.items()] + relaxable_categorical_inputs,  # type: ignore
        outputs=domain.outputs.features,  # type: ignore
        constraints=domain.constraints + new_constraints,
    )

    return new_domain, categorical_groups, relaxable_discrete_inputs  # type: ignore


def nchoosek_to_relaxable_domain_mapper(
    domain: Domain,
) -> Tuple[Domain, List[List[ContinuousInput]]]:
    var_occuring_in_nchoosek = []
    new_categories = []
    new_constraints = []
    n_choose_k_constraints = domain.constraints.get(includes=NChooseKConstraint)

    for constr in n_choose_k_constraints:
        var_occuring_in_nchoosek.extend(constr.features)  # type: ignore

        current_features: List[Feature] = [
            domain.get_feature(k) for k in constr.features  # type: ignore
        ]
        new_relaxable_categorical_vars, new_nchoosek_constraints = NChooseKGroup(
            current_features, constr.min_count, constr.max_count, constr.none_also_valid  # type: ignore
        )
        new_categories.append(new_relaxable_categorical_vars)
        new_constraints.extend(new_nchoosek_constraints)

        # allow vars to be set to 0
        for var in var_occuring_in_nchoosek:
            current_var = domain.inputs.get_by_key(var)
            if current_var.lower_bound > 0:  # type: ignore
                current_var.bounds = (0, current_var.upper_bound)  # type: ignore
            elif current_var.upper_bound < 0:  # type: ignore
                current_var.bounds = (current_var.lower_bound, 0)  # type: ignore

    new_domain = Domain(
        inputs=domain.inputs.features + [var for group in new_categories for var in group],  # type: ignore
        outputs=domain.outputs,
        constraints=domain.constraints.get(excludes=NChooseKConstraint)
        + new_constraints,
    )  # type: ignore
    return new_domain, new_categories


def NChooseKGroup_with_quantity(
    unique_group_identifier: str,
    keys: List[str],
    pick_at_least: int,
    pick_at_most: int,
    quantity_if_picked: Optional[
        Union[Tuple[float, float], List[Tuple[float, float]]]
    ] = None,
    combined_quantity_limit: Optional[float] = None,
    combined_quantity_is_equal_or_less_than: bool = False,
    use_non_relaxable_category_and_non_linear_constraint: bool = False,
) -> tuple[
    Union[List[CategoricalInput], List[ContinuousInput]],
    List[ContinuousInput],
    List[
        Union[
            LinearEqualityConstraint,
            LinearInequalityConstraint,
            NonlinearInequalityConstraint,
        ]
    ],
]:
    """
    helper function to generate an N choose K problem with categorical variables, with an option to connect each
    element of a category to a corresponding quantity of how much that category should be used.

    Args:
        unique_group_identifier (str): unique ID for the category/group which will be used to mark all variables
            containing to this group
        keys (List[str]): defines the names and the amount of the elements within the category
        pick_at_least (int): minimum number of elements to be picked from the category. >=0
        pick_at_most (int): maximum number of elements to be picked from the category. >=pick_at_least
        quantity_if_picked (Optional[Union[Tuple[float, float], List[Tuple[float, float]]]): If provided, specifies
            the lower and upper bound of the quantity, for each element in the category. List of bounds to specify the
            allowed quantity for each element separately or one single bound to set the same bounds for all elements.
        combined_quantity_limit (Optional[float]): If provided, sets an upper bound on what the sum of all the
            quantities of all elements should be
        combined_quantity_is_equal_or_less_than (bool): If True, the combined_quantity_limit describes the exact amount
            of the sum of all quantities. If False, it is a upper bound, i.e. the sum of the quantities can be lower.
            Default is False
        use_non_relaxable_category_and_non_linear_constraint (bool): Default is False.
            If False, ContinuousInput is used in combination with LinearConstraints as a relaxation.
            If True, CategoricalInput used in combination with NonlinearConstraints, as CategoricalInput can not be
            used within LinearConstraints
    Returns:
        Either one CategoricalInput wrapped in a List or List of ContinuousInput describing the group,
        If quantities are provided, List of ContinuousInput describing the quantity of each element of the group
        otherwise empty List,
        List of either LinearConstraints or mix of Linear- and NonlinearConstraints, which enforce the quantities
        and group restrictions.
    """
    if quantity_if_picked is not None:
        if type(quantity_if_picked) is list and len(keys) != len(quantity_if_picked):
            raise ValueError(
                f"number of keys must be the same as corresponding quantities. Received {len(keys)} keys "
                f"and {len(quantity_if_picked)} quantities"
            )

        if type(quantity_if_picked) is list and True in [
            0 in q for q in quantity_if_picked
        ]:
            raise ValueError(
                "If an element out of the group is chosen, the quantity with which it is used must be "
                "larger than 0"
            )

    if pick_at_least > pick_at_most:
        raise ValueError(
            f"your upper bound to pick an element should be larger your lower bound. "
            f"Currently: pick_at_least {pick_at_least} > pick_at_most {pick_at_most}"
        )

    if pick_at_least < 0:
        raise ValueError(
            f"you should at least pick 0 elements. Currently  pick_at_least = {pick_at_least}"
        )

    if pick_at_most > len(keys):
        raise ValueError(
            f"you can not pick more elements than are available. "
            f"Received pick_at_most {pick_at_most} > number of keys {len(keys)}"
        )

    if "pick_none" in keys:
        raise ValueError("pick_none is not allowed as a key")

    if True in ["_" in k for k in keys]:
        raise ValueError('"_" is not allowed as an character in the keys')

    if quantity_if_picked is not None and type(quantity_if_picked) != list:
        quantity_if_picked = [quantity_if_picked for k in keys]  # type: ignore

    quantity_var, all_new_constraints = [], []
    quantity_constraints_lb, quantity_constraints_ub = [], []
    max_quantity_constraint = None

    # creating possible combination of n choose k
    combined_keys_as_tuple = []
    if pick_at_most > 1:
        for i in range(max(2, pick_at_least), pick_at_most + 1):
            combined_keys_as_tuple.extend(list(combinations(keys, i)))
    if pick_at_least <= 1:
        combined_keys_as_tuple.extend([[k] for k in keys])

    combined_keys = ["_".join(w) for w in combined_keys_as_tuple]

    # generating the quantity variables and corresponding constraints
    if quantity_if_picked:
        (
            quantity_var,
            quantity_constraints_lb,
            quantity_constraints_ub,
            max_quantity_constraint,
        ) = _generate_quantity_var_constr(
            unique_group_identifier,
            keys,
            quantity_if_picked,
            combined_keys,
            combined_keys_as_tuple,
            use_non_relaxable_category_and_non_linear_constraint,
            combined_quantity_limit,
            combined_quantity_is_equal_or_less_than,
        )

    # allowing to pick none
    if pick_at_least == 0:
        combined_keys.append(unique_group_identifier + "_pick_none")

    # adding the new possible combinations to the list of keys
    keys = [unique_group_identifier + "_" + k for k in combined_keys]

    # choosing between CategoricalInput and ContinuousInput
    if use_non_relaxable_category_and_non_linear_constraint:
        category = [CategoricalInput(key=unique_group_identifier, categories=keys)]
        # if we use_legacy_class is true this constraint will be added by the discrete_to_relaxable_domain_mapper function
        pick_exactly_one_of_group_const = []
    else:
        category = [ContinuousInput(key=k, bounds=(0, 1)) for k in keys]
        pick_exactly_one_of_group_const = [
            LinearEqualityConstraint(
                features=list(keys), coefficients=[1 for k in keys], rhs=1
            )
        ]

    all_new_constraints = (
        pick_exactly_one_of_group_const
        + quantity_constraints_lb
        + quantity_constraints_ub
    )
    if max_quantity_constraint is not None:
        all_new_constraints.append(max_quantity_constraint)  # type: ignore
    return category, quantity_var, all_new_constraints


def _generate_quantity_var_constr(
    unique_group_identifier,
    keys,
    quantity_if_picked,
    combined_keys,
    combined_keys_as_tuple,
    use_non_relaxable_category_and_non_linear_constraint,
    combined_quantity_limit,
    combined_quantity_is_equal_or_less_than,
) -> Tuple[
    List[ContinuousInput],
    Union[List[NonlinearInequalityConstraint], List[LinearInequalityConstraint]],
    Union[List[NonlinearInequalityConstraint], List[LinearInequalityConstraint]],
    Optional[Union[LinearEqualityConstraint, LinearInequalityConstraint]],
]:
    """
    Internal helper function just to create the quantity variables and the corresponding constraints.
    """
    quantity_var = [
        ContinuousInput(
            key=unique_group_identifier + "_" + k + "_quantity", bounds=(0, q[1])
        )
        for k, q in zip(keys, quantity_if_picked)
    ]

    all_quantity_features = []
    for k in keys:
        all_quantity_features.append(
            [
                unique_group_identifier + "_" + state_key
                for state_key, state_tuple in zip(combined_keys, combined_keys_as_tuple)
                if k in state_tuple
            ]
        )

    if use_non_relaxable_category_and_non_linear_constraint:
        quantity_constraints_lb = [
            NonlinearInequalityConstraint(
                expression="".join(
                    ["-" + unique_group_identifier + "_" + k + "_quantity"]
                    + [f" + {q[0]} * {key_c}" for key_c in combi]
                ),
                features=[unique_group_identifier + "_" + k + "_quantity"] + combi,
            )
            for combi, k, q in zip(all_quantity_features, keys, quantity_if_picked)
            if len(combi) >= 1
        ]

        quantity_constraints_ub = [
            NonlinearInequalityConstraint(
                expression="".join(
                    [unique_group_identifier + "_" + k + "_quantity"]
                    + [f" - {q[1]} * {key_c}" for key_c in combi]
                ),
                features=[unique_group_identifier + "_" + k + "_quantity"] + combi,
            )
            for combi, k, q in zip(all_quantity_features, keys, quantity_if_picked)
            if len(combi) >= 1
        ]
    else:
        quantity_constraints_lb = [
            LinearInequalityConstraint(
                features=[unique_group_identifier + "_" + k + "_quantity"] + combi,
                coefficients=[-1.0] + [q[0] for i in range(len(combi))],
                rhs=0,
            )
            for combi, k, q in zip(all_quantity_features, keys, quantity_if_picked)
            if len(combi) >= 1
        ]

        quantity_constraints_ub = [
            LinearInequalityConstraint(
                features=[unique_group_identifier + "_" + k + "_quantity"] + combi,
                coefficients=[1.0] + [-q[1] for i in range(len(combi))],
                rhs=0,
            )
            for combi, k, q in zip(all_quantity_features, keys, quantity_if_picked)
            if len(combi) >= 1
        ]

    max_quantity_constraint = None
    if combined_quantity_limit is not None:
        if combined_quantity_is_equal_or_less_than:
            max_quantity_constraint = LinearEqualityConstraint(
                features=[q.key for q in quantity_var],
                coefficients=[1 for q in quantity_var],
                rhs=combined_quantity_limit,
            )
        else:
            max_quantity_constraint = LinearInequalityConstraint(
                features=[q.key for q in quantity_var],
                coefficients=[1 for q in quantity_var],
                rhs=combined_quantity_limit,
            )

    return (
        quantity_var,
        quantity_constraints_lb,
        quantity_constraints_ub,
        max_quantity_constraint,
    )


def NChooseKGroup(
    variables: List[Feature],
    pick_at_least: int,
    pick_at_most: int,
    none_also_valid: bool,
) -> Tuple[
    List[ContinuousInput],
    List[Union[LinearEqualityConstraint, LinearInequalityConstraint]],
]:
    """
    helper function to generate an N choose K problem with categorical variables, with an option to connect each
    element of a category to a corresponding quantity of how much that category should be used.

    Args:
        variables (List[ContinuousInput]): variables to pick from
        pick_at_least (int): minimum number of elements to be picked from the category. >=0
        pick_at_most (int): maximum number of elements to be picked from the category. >=pick_at_least
        none_also_valid (bool): defines if also none of the elements can be picked
    Returns:
        List of ContinuousInput describing the group,
        List of either LinearConstraints, which enforce the quantities
        and group restrictions.
    """

    keys = [var.key for var in variables]
    if pick_at_least > pick_at_most:
        raise ValueError(
            f"your upper bound to pick an element should be larger your lower bound. "
            f"Currently: pick_at_least {pick_at_least} > pick_at_most {pick_at_most}"
        )

    if pick_at_least < 0:
        raise ValueError(
            f"you should at least pick 0 elements. Currently  pick_at_least = {pick_at_least}"
        )

    if pick_at_most > len(keys):
        raise ValueError(
            f"you can not pick more elements than are available. "
            f"Received pick_at_most {pick_at_most} > number of keys {len(keys)}"
        )

    if "pick_none" in keys:
        raise ValueError("pick_none is not allowed as a key")

    # creating possible combination of n choose k
    combined_keys_as_tuple = []
    if pick_at_most > 1:
        for i in range(max(2, pick_at_least), pick_at_most + 1):
            combined_keys_as_tuple.extend(list(combinations(keys, i)))
    if pick_at_least <= 1:
        combined_keys_as_tuple.extend([[k] for k in keys])

    combined_keys = ["_".join(w) for w in combined_keys_as_tuple]
    combined_keys = ["categorical_helper" + "_" + k for k in combined_keys]

    # generating the corresponding constraints
    valid_states = []
    for k in keys:
        valid_states.append(
            [
                state_key
                for state_key, state_tuple in zip(combined_keys, combined_keys_as_tuple)
                if k in state_tuple
            ]
        )

    quantity_constraints_lb = [
        LinearInequalityConstraint(
            features=[var.key] + combi,
            coefficients=[-1.0] + [var.lower_bound for i in range(len(combi))],  # type: ignore
            rhs=0,
        )
        for combi, var in zip(valid_states, variables)
        if len(combi) >= 1
    ]

    quantity_constraints_ub = [
        LinearInequalityConstraint(
            features=[var.key] + combi,
            coefficients=[1.0] + [-var.upper_bound for i in range(len(combi))],  # type: ignore
            rhs=0,
        )
        for combi, var in zip(valid_states, variables)
        if len(combi) >= 1
    ]

    # allowing to pick none
    if pick_at_least == 0 or none_also_valid:
        combined_keys.append("categorical_helper_pick_none_of_" + "".join(keys))

    # adding the new possible combinations to the list of keys
    keys = combined_keys

    category = [ContinuousInput(key=k, bounds=(0, 1)) for k in keys]
    pick_exactly_one_of_group_const = [
        LinearEqualityConstraint(
            features=list(keys), coefficients=[1 for k in keys], rhs=1
        )
    ]

    all_new_constraints = []
    all_new_constraints.extend(pick_exactly_one_of_group_const)
    all_new_constraints.extend(quantity_constraints_lb)
    all_new_constraints.extend(quantity_constraints_ub)

    return category, all_new_constraints


def generate_mixture_constraints(
    keys: List[str],
) -> Tuple[LinearEqualityConstraint, List[ContinuousInput]]:
    binary_vars = (ContinuousInput(key=x, bounds=(0, 1)) for x in keys)

    mixture_constraint = LinearEqualityConstraint(
        features=keys, coefficients=[1 for x in range(len(keys))], rhs=1
    )

    return mixture_constraint, list(binary_vars)


def design_from_original_to_new_domain(
    original_domain: Domain, new_domain: Domain, design: pd.DataFrame
) -> pd.DataFrame:
    raise NotImplementedError(
        "mapping a design to a new domain is not implemented yet."
    )


def design_from_new_to_original_domain(
    original_domain: Domain, design: pd.DataFrame
) -> pd.DataFrame:
    # map the ContinuousInput describing the categoricals to the corresponding CategoricalInputs, choose random for multiple solutions
    transformed_design = design[
        original_domain.get_feature_keys(excludes=[CategoricalInput, Output])
    ]

    for group in original_domain.get_features(includes=CategoricalInput):
        categorical_columns = design[group.categories]  # type: ignore
        mask = ~np.isclose(categorical_columns.to_numpy(), 0)

        for i, row in enumerate(mask):
            non_zero_indices = np.argwhere(row)
            if non_zero_indices.size != 0:
                index_to_keep = np.random.choice(np.argwhere(row).flatten())
            else:
                index_to_keep = list(range(len(row)))
            mask[i] = np.zeros_like(row, dtype=bool)
            mask[i][index_to_keep] = True

        categorical_columns = categorical_columns.where(
            np.invert(mask),
            pd.DataFrame(
                np.full(
                    (len(categorical_columns), len(group.categories)),  # type: ignore
                    group.categories,  # type: ignore
                ),
                columns=categorical_columns.columns,
                index=categorical_columns.index,
            ),
        )
        categorical_columns = categorical_columns.where(
            mask,
            pd.DataFrame(
                np.full((len(categorical_columns), len(group.categories)), ""),  # type: ignore
                columns=categorical_columns.columns,
                index=categorical_columns.index,
            ),
        )
        transformed_design[group.key] = categorical_columns.apply("".join, axis=1)

    # map the ContinuousInput describing the discrete to the closest valid value
    for var in original_domain.get_features(includes=DiscreteInput):
        closest_solution = var.from_continuous(transformed_design)  # type: ignore
        transformed_design[var.key] = closest_solution

    return transformed_design


def equal_count_split(
    values: List[float], lower_bound: float, upper_bound: float
) -> Tuple[float, float]:
    """
    Determines the two elements x and y such that the intervals (lower_bound, x) and (y, upper_bound)
    have the same number of elements regarding the values of the discrete variable
    Args:
        values: the values to split into a range
        lower_bound: inclusive lower bound
        upper_bound: inclusive upper bound

    Returns: tuple of floats which split the interval in half

    """
    values.sort()
    sub_list = [elem for elem in values if lower_bound <= elem <= upper_bound]

    size = len(sub_list)
    if size % 2 == 0:
        lower_index = size / 2 - 1
        upper_index = size / 2
    elif size == 1:
        return sub_list[0], sub_list[0]
    else:
        lower_index = math.floor(size / 2)
        upper_index = math.ceil(size / 2)

    lower_index = int(lower_index)
    upper_index = int(upper_index)

    return sub_list[lower_index], sub_list[upper_index]



import itertools
from abc import ABC
from typing import List

import botorch
import pandas as pd
import torch
from botorch.models import ModelList
from botorch.models.transforms.input import ChainedInputTransform, FilterFeatures

from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import TInputTransformSpecs
from bofire.data_models.surrogates.api import BotorchSurrogates as DataModel
from bofire.surrogates.botorch import BotorchSurrogate
from bofire.surrogates.mapper import map as map_surrogate
from bofire.surrogates.trainable import TrainableSurrogate


class BotorchSurrogates(ABC):
    surrogates: List[BotorchSurrogate]

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        self.surrogates = [map_surrogate(model) for model in data_model.surrogates]  # type: ignore

    @property
    def input_preprocessing_specs(self) -> TInputTransformSpecs:
        return {
            key: value
            for model in self.surrogates
            for key, value in model.input_preprocessing_specs.items()
        }

    def fit(self, experiments: pd.DataFrame):
        for model in self.surrogates:
            if isinstance(model, TrainableSurrogate):
                model.fit(experiments)

    @property
    def outputs(self) -> Outputs:
        return Outputs(
            features=list(
                itertools.chain.from_iterable(
                    [model.outputs.get() for model in self.surrogates]  # type: ignore
                )
            )
        )

    # TODO: is this really neede here, code duplication with functional model
    def _check_compability(self, inputs: Inputs, outputs: Outputs):
        used_output_feature_keys = self.outputs.get_keys()
        if sorted(used_output_feature_keys) != sorted(outputs.get_keys()):
            raise ValueError("Output features do not match.")
        used_feature_keys = []
        for i, model in enumerate(self.surrogates):
            if len(model.inputs) > len(inputs):
                raise ValueError(
                    f"Model with index {i} has more features than acceptable."
                )
            for feat in model.inputs:
                try:
                    other_feat = inputs.get_by_key(feat.key)
                except KeyError:
                    raise ValueError(f"Feature {feat.key} not found.")
                # now compare the features
                # TODO: make more sohisticated comparisons based on the type
                # has to to be implemented in features, for the start
                # we go with __eq__
                if feat != other_feat:
                    raise ValueError(f"Features with key {feat.key} are incompatible.")
                if feat.key not in used_feature_keys:
                    used_feature_keys.append(feat.key)
        if len(used_feature_keys) != len(inputs):
            raise ValueError("Unused features are present.")

    def compatibilize(self, inputs: Inputs, outputs: Outputs) -> ModelList:
        # TODO: add sync option
        # check if models are compatible to provided inputs and outputs
        # of the optimization domain
        self._check_compability(inputs=inputs, outputs=outputs)
        features2idx, _ = inputs._get_transform_info(self.input_preprocessing_specs)
        #
        all_gp = True
        botorch_models = []
        # we sort the models by sorting them with their occurence in outputs
        for output_feature_key in outputs.get_keys():
            # get the corresponding model
            model = {model.outputs[0].key: model for model in self.surrogates}[
                output_feature_key
            ]
            if model.model is None:
                raise ValueError(
                    f"Surrogate for output feature {output_feature_key} not fitted."
                )
            # in case that inputs are complete we do not need to adjust anything
            if len(model.inputs) == len(inputs):
                botorch_models.append(model.model)
            # in this case we have to care for the indices
            if len(model.inputs) < len(inputs):
                indices = []
                for key in model.inputs.get_keys():
                    indices += features2idx[key]
                features_filter = FilterFeatures(
                    feature_indices=torch.tensor(indices, dtype=torch.int64),
                    transform_on_train=False,
                )
                if (
                    hasattr(model.model, "input_transform")
                    and model.model.input_transform is not None  # type: ignore
                ):
                    model.model.input_transform = ChainedInputTransform(  # type: ignore
                        tcompatibilize=features_filter, tf2=model.model.input_transform  # type: ignore
                    )
                else:
                    model.model.input_transform = features_filter  # type: ignore

                botorch_models.append(model.model)
            if isinstance(model.model, botorch.models.SingleTaskGP) is False:
                all_gp = False

        if len(botorch_models) == 1:
            return botorch_models[0]
        if all_gp:
            return botorch.models.ModelListGP(*botorch_models)
        return ModelList(*botorch_models)



import importlib.util

import numpy as np
import pandas as pd
import pytest

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    ContinuousInput,
    ContinuousOutput,
)
from bofire.strategies.doe.design import (
    check_fixed_experiments,
    check_partially_and_fully_fixed_experiments,
    find_local_max_ipopt,
    get_n_experiments,
)
from bofire.strategies.doe.utils import get_formula_from_string, n_zero_eigvals

CYIPOPT_AVAILABLE = importlib.util.find_spec("cyipopt") is not None


@pytest.mark.skipif(CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_raise_error_if_cyipopt_not_available():
    pytest.raises(ImportError)


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_no_constraint():
    # Design for a problem with an n-choose-k constraint
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(4)
        ],
        outputs=[ContinuousOutput(key="y")],
    )
    dim_input = len(domain.inputs.get_keys())

    num_exp = (
        len(get_formula_from_string(model_type="linear", domain=domain))
        - n_zero_eigvals(domain=domain, model_type="linear")
        + 3
    )

    design = find_local_max_ipopt(domain, "linear")
    assert design.shape == (num_exp, dim_input)


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_nchoosek():
    # Design for a problem with an n-choose-k constraint
    inputs = [
        ContinuousInput(
            key=f"x{i+1}",
            bounds=(0, 1),
        )
        for i in range(4)
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=[f"x{i+1}" for i in range(4)],
                min_count=0,
                max_count=3,
                none_also_valid=True,
            )
        ],
    )

    D = len(domain.inputs)

    N = (
        len(get_formula_from_string(model_type="linear", domain=domain))
        - n_zero_eigvals(domain=domain, model_type="linear")
        + 3
    )
    print(N)

    A = find_local_max_ipopt(domain, "linear")
    assert A.shape == (N, D)


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_mixture():
    # Design for a problem with a mixture constraint
    inputs = [
        ContinuousInput(
            key=f"x{i+1}",
            bounds=(0, 1),
        )
        for i in range(4)
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=[f"x{i+1}" for i in range(4)], coefficients=[1, 1, 1, 1], rhs=1
            )
        ],
    )

    D = len(domain.inputs)

    N = len(get_formula_from_string(domain=domain, model_type="linear")) + 3
    A = find_local_max_ipopt(domain, "linear")
    assert A.shape == (N, D)


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_mixed_results():
    inputs = [
        ContinuousInput(
            key=f"x{1}",
            bounds=(0, 1),
        ),
        ContinuousInput(
            key=f"x{2}",
            bounds=(0, 1),
        ),
        ContinuousInput(
            key=f"x{3}",
            bounds=(0, 1),
        ),
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
            ),
            NChooseKConstraint(
                features=[f"x{i+1}" for i in range(3)],
                min_count=0,
                max_count=1,
                none_also_valid=True,
            ),
        ],
    )

    # with pytest.warns(ValueError):
    A = find_local_max_ipopt(domain, "fully-quadratic", ipopt_options={"maxiter": 100})
    opt = np.eye(3)
    for row in A.to_numpy():
        assert any(np.allclose(row, o, atol=1e-2) for o in opt)
    for o in opt:
        assert any(np.allclose(o, row, atol=1e-2) for row in A.to_numpy())


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_results():
    # define problem: no NChooseK constraints
    inputs = [
        ContinuousInput(
            key=f"x{1}",
            bounds=(0, 1),
        ),
        ContinuousInput(key=f"x{2}", bounds=(0.1, 1)),
        ContinuousInput(key=f"x{3}", bounds=(0, 0.6)),
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
            ),
            LinearInequalityConstraint(
                features=["x1", "x2"], coefficients=[5, 4], rhs=3.9
            ),
            LinearInequalityConstraint(
                features=["x1", "x2"], coefficients=[-20, 5], rhs=-3
            ),
        ],
    )
    np.random.seed(1)
    A = find_local_max_ipopt(domain, "linear", n_experiments=12)
    opt = np.array([[0.2, 0.2, 0.6], [0.3, 0.6, 0.1], [0.7, 0.1, 0.2], [0.3, 0.1, 0.6]])
    for row in A.to_numpy():
        assert any(np.allclose(row, o, atol=1e-2) for o in opt)
    for o in opt[:-1]:
        assert any(np.allclose(o, row, atol=1e-2) for row in A.to_numpy())


# def test_find_local_max_ipopt_sampling():
#     # define problem
#     problem = opti.Problem(
#         inputs=[opti.Continuous(f"x{i}", [0, 1]) for i in range(3)],
#         outputs=[opti.Continuous("y")],
#     )


#     # test sampling methods
#     find_local_max_ipopt(problem, "linear", sampling=OptiSampling)
#     find_local_max_ipopt(problem, "linear", sampling=CornerSampling)
#     find_local_max_ipopt(problem, "linear", sampling=ProbabilitySimplexSampling)
#     sampling = np.zeros(shape=(10, 3)).flatten()
#     find_local_max_ipopt(problem, "linear", n_experiments=10, sampling=sampling)


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_fixed_experiments():
    # TODO fix this test. Currently it gets stuck in a local minimum 50% of the time
    # working if fixed experiments is tested below. Do I need this test?
    inputs = [
        ContinuousInput(
            key=f"x{1}",
            bounds=(0, 1),
        ),
        ContinuousInput(key=f"x{2}", bounds=(0.1, 1)),
        ContinuousInput(key=f"x{3}", bounds=(0, 0.6)),
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
            ),
            LinearInequalityConstraint(
                features=["x1", "x2"], coefficients=[5, 4], rhs=3.9
            ),
            LinearInequalityConstraint(
                features=["x1", "x2"], coefficients=[-20, 5], rhs=-3
            ),
        ],
    )
    # np.random.seed(1)
    # fixed_experiments = pd.DataFrame([[0.3, 0.5, 0.2]], columns=["x1", "x2", "x3"])
    # A = find_local_max_ipopt(
    #     domain,
    #     "linear",
    #     n_experiments=12,
    #     fixed_experiments=fixed_experiments,  # type: ignore
    # )
    # opt = np.array(
    #     [
    #         [0.2, 0.2, 0.6],
    #         [0.3, 0.6, 0.1],
    #         [0.7, 0.1, 0.2],
    #         [0.3, 0.1, 0.6],
    #         [0.3, 0.5, 0.2],
    #     ]
    # )
    # for row in A.to_numpy():
    #     assert any([np.allclose(row, o, atol=1e-2) for o in opt])
    # for o in opt[:-1]:
    #     assert any([np.allclose(o, row, atol=1e-2) for row in A.to_numpy()])
    # assert np.allclose(A.to_numpy()[0, :], np.array([0.3, 0.5, 0.2]))

    # define domain: no NChooseK constraints, invalid proposal
    np.random.seed(1)
    with pytest.raises(ValueError):
        find_local_max_ipopt(
            domain,
            "linear",
            n_experiments=12,
            fixed_experiments=pd.DataFrame(
                np.ones(shape=(12, 3)), columns=["x1", "x2", "x3"]
            ),
        )

    # define domain: with NChooseK constraints, 2 fixed_experiments
    inputs = [
        ContinuousInput(
            key=f"x{1}",
            bounds=(0, 1),
        ),
        ContinuousInput(
            key=f"x{2}",
            bounds=(0, 1),
        ),
        ContinuousInput(
            key=f"x{3}",
            bounds=(0, 1),
        ),
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
            ),
            NChooseKConstraint(
                features=[f"x{i+1}" for i in range(3)],
                min_count=0,
                max_count=1,
                none_also_valid=True,
            ),
        ],
    )

    # with pytest.warns(ValueError):
    np.random.seed(1)
    A = find_local_max_ipopt(
        domain,
        "fully-quadratic",
        ipopt_options={"maxiter": 100},
        fixed_experiments=pd.DataFrame([[1, 0, 0], [0, 1, 0]], columns=["x1", "x2", "x3"]),  # type: ignore
    )
    opt = np.eye(3)
    for row in A.to_numpy():
        assert any(np.allclose(row, o, atol=1e-2) for o in opt)
    for o in opt:
        assert any(np.allclose(o, row, atol=1e-2) for row in A.to_numpy())
    assert np.allclose(A.to_numpy()[:2, :], opt[:2, :])


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_check_fixed_experiments():
    # define problem: everything fine
    inputs = [
        ContinuousInput(
            key=f"x{1}",
            bounds=(0, 1),
        ),
        ContinuousInput(
            key=f"x{2}",
            bounds=(0, 1),
        ),
        ContinuousInput(
            key=f"x{3}",
            bounds=(0, 1),
        ),
    ]
    domain = Domain.from_lists(
        inputs=inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            LinearEqualityConstraint(
                features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
            ),
            NChooseKConstraint(
                features=[f"x{i+1}" for i in range(3)],
                min_count=0,
                max_count=1,
                none_also_valid=True,
            ),
        ],
    )
    fixed_experiments = pd.DataFrame(
        np.array([[1, 0, 0], [0, 1, 0]]), columns=domain.inputs.get_keys()
    )
    check_fixed_experiments(domain, 3, fixed_experiments)

    # define problem: not enough experiments
    fixed_experiments = pd.DataFrame(
        np.array([[1, 0, 0], [0, 1, 0]]), columns=domain.inputs.get_keys()
    )
    with pytest.raises(ValueError):
        check_fixed_experiments(domain, 2, fixed_experiments)

    # # define problem: invalid shape
    # fixed_experiments = pd.DataFrame(
    #     np.array([[1, 0, 0, 0], [0, 1, 0, 0]]), columns=domain.inputs.get_keys()
    # )
    # with pytest.raises(ValueError):
    #     check_fixed_experiments(domain, 3, fixed_experiments)


# @pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
# def test_check_constraints_and_domain_respected():
#     # problem with unfulfillable constraints
#     # formulation constraint
#     inputs = [
#         ContinuousInput(key=f"x{1}", lower_bound=0.5, upper_bound=1),
#         ContinuousInput(key=f"x{2}", lower_bound=0.5, upper_bound=1),
#         ContinuousInput(key=f"x{3}", lower_bound=0.5, upper_bound=1),
#     ]
#     domain = Domain(
#         inputs=inputs,
#         outputs=[ContinuousOutput(key="y")],
#         constraints=[
#             LinearEqualityConstraint(
#                 features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
#             ),
#         ],
#     )

#     with warnings.catch_warnings():
#         # warnings.simplefilter("ignore")
#         try:
#             A = find_local_max_ipopt(domain=domain, model_type="linear")
#         except Exception as e:
#             assert (
#                 str(e)
#                 == "No feasible point found. Constraint polytope appears empty. Check your constraints."
#             )

#     # with pytest.warns(UserWarning, match="Please check your results"):
#     #    domain.validate_candidates(candidates=A, only_inputs=True)

#     # everything ok
#     inputs = [
#         ContinuousInput(key=f"x{1}", bounds=(0, 1),),
#         ContinuousInput(key=f"x{2}", bounds=(0, 1),),
#         ContinuousInput(key=f"x{3}", bounds=(0, 1),),
#     ]
#     domain = Domain(
#         inputs=inputs,
#         outputs=[ContinuousOutput(key="y")],
#         constraints=[
#             LinearEqualityConstraint(
#                 features=[f"x{i+1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
#             ),
#             NChooseKConstraint(
#                 features=[f"x{i+1}" for i in range(3)],
#                 min_count=0,
#                 max_count=1,
#                 none_also_valid=True,
#             ),
#         ],
#     )

#     with warnings.catch_warnings():
#         warnings.simplefilter("ignore")
#         A = find_local_max_ipopt(domain=domain, model_type="linear")

#     with warnings.catch_warnings():
#         warnings.simplefilter("error")
#         domain.validate_candidates(candidates=A, only_inputs=True)


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_find_local_max_ipopt_nonlinear_constraint():
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key="x1", bounds=(-1, 1)),
            ContinuousInput(key="x2", bounds=(-1, 1)),
            ContinuousInput(key="x3", bounds=(0, 1)),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NonlinearInequalityConstraint(
                expression="x1**2 + x2**2 - x3",
                features=["x1", "x2", "x3"],
                jacobian_expression="[2*x1,2*x2,-1]",
            )
        ],
    )

    result = find_local_max_ipopt(domain, "linear", ipopt_options={"maxiter": 100})

    assert np.allclose(domain.constraints(result), 0, atol=1e-6)


def test_get_n_experiments():
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(key="x1", bounds=(-1, 1)),
            ContinuousInput(key="x2", bounds=(-1, 1)),
            ContinuousInput(key="x3", bounds=(0, 1)),
        ],
        outputs=[ContinuousOutput(key="y")],
    )

    # keyword
    assert get_n_experiments(domain, "linear") == 7

    # explicit formula
    assert get_n_experiments(domain, "x1 + x2 + x3 + x1:x2 + {x2**2}") == 9

    # user provided n_experiment
    with pytest.warns(UserWarning):
        assert get_n_experiments(domain, "linear", 4) == 4


@pytest.mark.skipif(not CYIPOPT_AVAILABLE, reason="requires cyipopt")
def test_partially_fixed_experiments():
    domain = Domain(
        inputs=[
            ContinuousInput(key="x1", bounds=(0, 5)),
            ContinuousInput(key="x2", bounds=(0, 15)),
            ContinuousInput(key="a1", bounds=(0, 1)),
            ContinuousInput(key="a2", bounds=(0, 1)),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            # Case 1: a and b are active
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, 1, 10, -10], rhs=15
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, 0.2, 2, -2], rhs=5
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, -1, -3, 3], rhs=5
            ),
            # Case 2: a and c are active
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, 1, -10, -10], rhs=5
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, 0.2, 2, 2], rhs=7
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, -1, -3, -3], rhs=2
            ),
            # Case 3: c and b are active
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, 1, 0, -10], rhs=5
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, 0.2, 0, 2], rhs=5
            ),
            LinearInequalityConstraint(
                features=["x1", "x2", "a1", "a2"], coefficients=[1, -1, 0, 3], rhs=5
            ),
        ],
    )
    fixed_experiments = pd.DataFrame(
        np.array([[1, 0, 0, 0], [0, 1, 0, 0]]), columns=domain.inputs.get_keys()
    )
    partially_fixed_experiments = pd.DataFrame(
        np.array([[1, None, None, None], [0, 1, 0, 0]]),
        columns=domain.inputs.get_keys(),
    )
    # all fine
    check_partially_and_fully_fixed_experiments(
        domain, 10, fixed_experiments, partially_fixed_experiments
    )

    # all fine
    check_partially_and_fully_fixed_experiments(
        domain, 4, fixed_experiments, partially_fixed_experiments
    )

    # partially fixed will be cut of
    with pytest.warns(UserWarning):
        check_partially_and_fully_fixed_experiments(
            domain, 3, fixed_experiments, partially_fixed_experiments
        )

    # to few experiments
    with pytest.raises(ValueError):
        check_partially_and_fully_fixed_experiments(
            domain, 2, fixed_experiments, partially_fixed_experiments
        )



from typing import Union

from bofire.data_models.features.categorical import CategoricalInput, CategoricalOutput
from bofire.data_models.features.continuous import ContinuousInput, ContinuousOutput
from bofire.data_models.features.descriptor import (
    CategoricalDescriptorInput,
    ContinuousDescriptorInput,
)
from bofire.data_models.features.discrete import DiscreteInput
from bofire.data_models.features.feature import (  # noqa: F401
    _CAT_SEP,
    Feature,
    Input,
    Output,
    TInputTransformSpecs,
)
from bofire.data_models.features.molecular import (
    CategoricalMolecularInput,
    MolecularInput,
)
from bofire.data_models.features.numerical import NumericalInput

AbstractFeature = Union[
    Feature,
    Input,
    Output,
    NumericalInput,
]

AnyFeature = Union[
    DiscreteInput,
    CategoricalInput,
    ContinuousOutput,
    ContinuousInput,
    ContinuousDescriptorInput,
    CategoricalDescriptorInput,
    MolecularInput,
    CategoricalOutput,
    CategoricalMolecularInput,
]

AnyInput = Union[
    DiscreteInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousDescriptorInput,
    CategoricalDescriptorInput,
    MolecularInput,
    CategoricalMolecularInput,
]

AnyOutput = Union[ContinuousOutput, CategoricalOutput]



import warnings

import numpy as np
import pandas as pd
import torch
from botorch.utils.sampling import get_polytope_samples

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    DiscreteInput,
)
from bofire.data_models.strategies.api import PolytopeSampler as DataModel
from bofire.strategies.samplers.sampler import SamplerStrategy
from bofire.utils.torch_tools import get_linear_constraints, tkwargs


class PolytopeSampler(SamplerStrategy):
    """Sampler that generates samples from a Polytope defined by linear equality and ineqality constraints.

    Attributes:
        domain (Domain): Domain defining the constrained input space
        fallback_sampling_method: SamplingMethodEnum, optional): Method to use for sampling when no
            constraints are present. Defaults to UNIFORM.
    """

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.fallback_sampling_method = data_model.fallback_sampling_method

    def _ask(self, n: int) -> pd.DataFrame:
        if len(self.domain.constraints) == 0:
            return self.domain.inputs.sample(n, self.fallback_sampling_method)

        # check if we have pseudo fixed features in the linear equality constraints
        # a pseudo fixed is a linear euquality constraint with only one feature included
        # this can happen when fixing features when sampling with NChooseK constraints
        eqs = get_linear_constraints(
            domain=self.domain,
            constraint=LinearEqualityConstraint,  # type: ignore
            unit_scaled=False,
        )
        cleaned_eqs = []
        pseudo_fixed = {}
        for eq in eqs:
            if (
                len(eq[0]) == 1
            ):  # only one coefficient, so this is a pseudo fixed feature
                pseudo_fixed[
                    self.domain.inputs.get_keys(ContinuousInput)[eq[0][0]]
                ] = float(eq[2] / eq[1][0])
            else:
                cleaned_eqs.append(eq)

        # we have to map the indices in case of fixed features
        # as we remove all fixed feature for the sampler, we have to adjust the
        # indices in the constraints, here we get the mapper to map original
        # to adjusted indices
        feature_map = {}
        counter = 0
        for i, feat in enumerate(self.domain.get_features(ContinuousInput)):
            if (not feat.is_fixed()) and (feat.key not in pseudo_fixed.keys()):  # type: ignore
                feature_map[i] = counter
                counter += 1

        # get the bounds
        lower = [
            feat.lower_bound  # type: ignore
            for feat in self.domain.get_features(ContinuousInput)
            if not feat.is_fixed() and feat.key not in pseudo_fixed.keys()  # type: ignore
        ]
        if len(lower) == 0:
            warnings.warn(
                "Nothing to sample, all is fixed. Just the fixed set is returned.",
                UserWarning,
            )
            samples = pd.DataFrame(
                data=np.nan, index=range(n), columns=self.domain.inputs.get_keys()
            )
        else:
            upper = [
                feat.upper_bound  # type: ignore
                for feat in self.domain.get_features(ContinuousInput)
                if not feat.is_fixed() and feat.key not in pseudo_fixed.keys()  # type: ignore
            ]
            bounds = torch.tensor([lower, upper]).to(**tkwargs)
            assert bounds.shape[-1] == len(feature_map) == counter

            # get the inequality constraints and map features back
            # we also check that only features present in the mapper
            # are present in the constraints
            ineqs = get_linear_constraints(
                domain=self.domain,
                constraint=LinearInequalityConstraint,  # type: ignore
                unit_scaled=False,
            )
            for ineq in ineqs:
                for key, value in feature_map.items():
                    if key != value:
                        ineq[0][ineq[0] == key] = value
                assert (
                    ineq[0].max() <= counter
                ), "Something went wrong when transforming the linear constraints. Revisit the problem."

            # map the indice of the equality constraints
            for eq in cleaned_eqs:
                for key, value in feature_map.items():
                    if key != value:
                        eq[0][eq[0] == key] = value
                assert (
                    eq[0].max() <= counter
                ), "Something went wrong when transforming the linear constraints. Revisit the problem."

            # now use the hit and run sampler
            candidates = get_polytope_samples(
                n=n,
                bounds=bounds.to(**tkwargs),
                inequality_constraints=ineqs if len(ineqs) > 0 else None,
                equality_constraints=cleaned_eqs if len(cleaned_eqs) > 0 else None,
                n_burnin=1000,
                # thinning=200
            )

            # check that the random generated candidates are not always the same
            if (candidates.unique(dim=0).shape[0] != n) and (n > 1):
                warnings.warn("Generated candidates are not unique!")

            free_continuals = [
                feat.key
                for feat in self.domain.get_features(ContinuousInput)
                if not feat.is_fixed() and feat.key not in pseudo_fixed.keys()  # type: ignore
            ]

            # setup the output
            samples = pd.DataFrame(
                data=candidates.detach().numpy().reshape(n, len(free_continuals)),
                index=range(n),
                columns=free_continuals,
            )

        # setup the categoricals and discrete ones as uniform sampled vals
        for feat in self.domain.get_features([CategoricalInput, DiscreteInput]):
            samples[feat.key] = feat.sample(n)  # type: ignore

        # setup the fixed continuous ones
        for feat in self.domain.inputs.get_fixed():
            samples[feat.key] = feat.fixed_value()[0]  # type: ignore

        # setup the pseudo fixed ones
        for key, value in pseudo_fixed.items():
            samples[key] = value

        return samples

    def duplicate(self, domain: Domain) -> SamplerStrategy:
        data_model = DataModel(
            domain=domain,
            fallback_sampling_method=self.fallback_sampling_method,
        )
        return self.__class__(data_model=data_model)



import math
from abc import abstractmethod
from copy import deepcopy

import pandas as pd
from pydantic import validate_arguments

from bofire.data_models.constraints.api import NChooseKConstraint
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import ContinuousInput
from bofire.data_models.strategies.api import SamplerStrategy as DataModel
from bofire.strategies.strategy import Strategy


class SamplerStrategy(Strategy):
    """Base class for sampling methods in BoFire for sampling from constrained input spaces.

    Attributes
        domain (Domain): Domain defining the constrained input space
    """

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    @abstractmethod
    def duplicate(self, domain: Domain) -> "SamplerStrategy":
        pass

    @validate_arguments
    def ask(
        self,
        candidate_count: int,
        return_all: bool = False,
        raise_validation_error: bool = True,
    ) -> pd.DataFrame:
        """Generates the samples. In the case that `NChooseK` constraints are
        present, per combination `n` samples are generated.

        Args:
            n (Tnum_asks): number of samples to generate.
            return_all (bool, optional): If true all `NchooseK` samples
                are generated, else only `n` samples in total. Defaults to True.
            raise_validation_error (bool, optional): If true an error will be raised if candidates violate constraints,
                otherwise only a warning will be displayed. Defaults to True.

        Returns:
            Dataframe with samples.
        """
        # n = candidate_count
        # handle here NChooseK
        if len(self.domain.constraints.get(NChooseKConstraint)) > 0:
            _, unused = self.domain.get_nchoosek_combinations()

            if return_all:
                sampled_combinations = unused
                num_samples_per_it = candidate_count
            else:
                if candidate_count <= len(unused):
                    sampled_combinations = [
                        unused[i]
                        for i in self.rng.choice(
                            len(unused), size=candidate_count, replace=False
                        )
                    ]
                    num_samples_per_it = 1
                else:
                    sampled_combinations = unused
                    num_samples_per_it = math.ceil(candidate_count / len(unused))

            samples = []
            for u in sampled_combinations:
                # create new domain without the nchoosekconstraints
                domain = deepcopy(self.domain)
                domain.constraints = domain.constraints.get(excludes=NChooseKConstraint)
                # fix the unused features
                for key in u:
                    feat = domain.inputs.get_by_key(key=key)
                    assert isinstance(feat, ContinuousInput)
                    feat.bounds = (0, 0)
                # setup then sampler for this situation
                sampler: SamplerStrategy = self.duplicate(domain=domain)
                samples.append(sampler.ask(num_samples_per_it))
            samples = pd.concat(samples, axis=0, ignore_index=True)
            if return_all:
                return self.domain.validate_candidates(
                    samples,
                    only_inputs=True,
                    raise_validation_error=raise_validation_error,
                )
            return self.domain.validate_candidates(
                samples.sample(n=candidate_count, replace=False, ignore_index=True),
                only_inputs=True,
                raise_validation_error=raise_validation_error,
            )

        samples = self._ask(candidate_count)
        if len(samples) != candidate_count:
            raise ValueError(f"expected {candidate_count} samples, got {len(samples)}.")
        return self.domain.validate_candidates(
            samples, only_inputs=True, raise_validation_error=raise_validation_error
        )

    def has_sufficient_experiments(self) -> bool:
        return True



from abc import abstractmethod
from typing import List, Optional

import pandas as pd
from pydantic import Field
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel


class Constraint(BaseModel):
    """Abstract base class to define constraints on the optimization space."""

    type: str

    @abstractmethod
    def is_fulfilled(
        self, experiments: pd.DataFrame, tol: Optional[float] = 1e-6
    ) -> pd.Series:
        """Abstract method to check if a constraint is fulfilled for all the rows of the provided dataframe.

        Args:
            experiments (pd.DataFrame): Dataframe to check constraint fulfillment.
            tol (float, optional): tolerance parameter. A constraint is considered as not fulfilled if
                the violation is larger than tol. Defaults to 0.

        Returns:
            bool: True if fulfilled else False
        """
        pass

    @abstractmethod
    def __call__(self, experiments: pd.DataFrame) -> pd.Series:
        """Numerically evaluates the constraint.

        Args:
            experiments (pd.DataFrame): Dataframe to evaluate the constraint on.

        Returns:
            pd.Series: Distance to reach constraint fulfillment.
        """
        pass

    @abstractmethod
    def jacobian(self, experiments: pd.DataFrame) -> pd.DataFrame:
        """Numerically evaluates the jacobian of the constraint
        Args:
            experiments (pd.DataFrame): Dataframe to evaluate the constraint on.
        Returns:
            pd.DataFrame: the i-th row contains the jacobian evaluated at the i-th experiment
        """
        pass


class ConstraintError(Exception):
    """Base Error for Constraints"""

    pass


class ConstraintNotFulfilledError(ConstraintError):
    """Raised when an constraint is not fulfilled."""

    pass


FeatureKeys = Annotated[List[str], Field(min_items=2)]
Coefficients = Annotated[List[float], Field(min_items=2)]



import collections.abc
from itertools import chain
from typing import List, Literal, Sequence, Type, Union

import pandas as pd
from pydantic import Field

from bofire.data_models.base import BaseModel, filter_by_class
from bofire.data_models.constraints.api import AnyConstraint, Constraint


class Constraints(BaseModel):
    type: Literal["Constraints"] = "Constraints"
    constraints: Sequence[AnyConstraint] = Field(default_factory=lambda: [])

    def __iter__(self):
        return iter(self.constraints)

    def __len__(self):
        return len(self.constraints)

    def __getitem__(self, i):
        return self.constraints[i]

    def __add__(
        self, other: Union[Sequence[AnyConstraint], "Constraints"]
    ) -> "Constraints":
        if isinstance(other, collections.abc.Sequence):
            other_constraints = other
        else:
            other_constraints = other.constraints
        constraints = list(chain(self.constraints, other_constraints))
        return Constraints(constraints=constraints)

    def __call__(self, experiments: pd.DataFrame) -> pd.DataFrame:
        """Numerically evaluate all constraints

        Args:
            experiments (pd.DataFrame): data to evaluate the constraint on

        Returns:
            pd.DataFrame: Constraint evaluation for each of the constraints
        """
        return pd.concat([c(experiments) for c in self.constraints], axis=1)

    def jacobian(self, experiments: pd.DataFrame) -> list:
        """Numerically evaluate the jacobians of all constraints

        Args:
            experiments (pd.DataFrame): data to evaluate the constraint jacobians on

        Returns:
            list: A list containing the jacobians as pd.DataFrames
        """
        return [c.jacobian(experiments) for c in self.constraints]

    def is_fulfilled(self, experiments: pd.DataFrame, tol: float = 1e-6) -> pd.Series:
        """Check if all constraints are fulfilled on all rows of the provided dataframe

        Args:
            experiments (pd.DataFrame): Dataframe with data, the constraint validity should be tested on
            tol (float, optional): tolerance parameter. A constraint is considered as not fulfilled if
                the violation is larger than tol. Defaults to 0.

        Returns:
            Boolean: True if all constraints are fulfilled for all rows, false if not
        """
        if len(self.constraints) == 0:
            return pd.Series([True] * len(experiments), index=experiments.index)
        return pd.concat(
            [c.is_fulfilled(experiments, tol) for c in self.constraints], axis=1
        ).all(axis=1)

    def get(
        self,
        includes: Union[Type, List[Type]] = Constraint,
        excludes: Union[Type, List[Type]] = None,
        exact: bool = False,
    ) -> "Constraints":
        """get constraints of the domain

        Args:
            includes (Union[Constraint, List[Constraint]], optional): Constraint class or list of specific constraint classes to be returned. Defaults to Constraint.
            excludes (Union[Type, List[Type]], optional): Constraint class or list of specific constraint classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact class listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.

        Returns:
            List[Constraint]: List of constraints in the domain fitting to the passed requirements.
        """
        return Constraints(
            constraints=filter_by_class(
                self.constraints,
                includes=includes,
                excludes=excludes,
                exact=exact,
            )
        )



import itertools
import unittest
import warnings
from typing import Literal, Type

import pandas as pd
import pytest
import torch
from botorch.optim.initializers import gen_batch_initial_conditions

import bofire.data_models.strategies.api as data_models
import bofire.data_models.surrogates.api as surrogate_data_models
import bofire.strategies.api as strategies
import tests.bofire.data_models.specs.api as specs
from bofire.benchmarks.single import Hartmann
from bofire.data_models.constraints.api import (
    Constraint,
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.enum import CategoricalEncodingEnum, CategoricalMethodEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Feature,
    Output,
)
from bofire.data_models.objectives.api import MaximizeObjective, MinimizeObjective
from bofire.utils.torch_tools import get_nchoosek_constraints, tkwargs
from tests.bofire.data_models.test_domain_validators import generate_experiments
from tests.bofire.strategies.specs import (
    VALID_ALLOWED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
    VALID_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
    VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
    VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
    VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
    VALID_DISCRETE_INPUT_FEATURE_SPEC,
    VALID_FIXED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
    VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC,
    VALID_FIXED_CONTINUOUS_INPUT_FEATURE_SPEC,
)

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning, append=True)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")


class DummyFeature(Feature):
    type: Literal["DummyFeature"] = "DummyFeature"

    def is_fixed(self):
        pass

    def fixed_value(self):
        pass


class DummyStrategyDataModel(data_models.BotorchStrategy):
    type: Literal["DummyStrategyDataModel"] = "DummyStrategyDataModel"

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        return my_type in [
            LinearEqualityConstraint,
            LinearInequalityConstraint,
            NChooseKConstraint,
        ]

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [
            ContinuousInput,
            CategoricalInput,
            DiscreteInput,
            CategoricalDescriptorInput,
            ContinuousOutput,
        ]

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Feature]) -> bool:
        return my_type in [MinimizeObjective, MaximizeObjective]


class DummyStrategy(strategies.BotorchStrategy):
    def _get_acqfs(
        self,
    ) -> None:
        pass


if1 = ContinuousInput(
    **{
        **VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if1",
    }
)
if2 = ContinuousInput(
    **{
        **VALID_FIXED_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if2",
    }
)

if3 = CategoricalInput(
    **{
        **VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
        "key": "if3",
    }
)

if4 = CategoricalInput(
    **{
        **VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC,
        "key": "if4",
    }
)

if5 = CategoricalDescriptorInput(
    **{
        **VALID_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
        "key": "if5",
    }
)

if6 = CategoricalDescriptorInput(
    **{
        **VALID_FIXED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
        "key": "if6",
    }
)

if7 = DummyFeature(key="if7")

if8 = CategoricalDescriptorInput(
    **{
        **VALID_ALLOWED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
        "key": "if8",
    }
)

if9 = DiscreteInput(
    **{
        **VALID_DISCRETE_INPUT_FEATURE_SPEC,
        "key": "if9",
    }
)

of1 = ContinuousOutput(
    **{
        **VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
        "key": "of1",
    }  # type: ignore
)

of2 = ContinuousOutput(
    **{
        **VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
        "key": "of2",
    }  # type: ignore
)

domains = [
    Domain.from_lists(
        inputs=[if1, if3, if5, if9],  # no fixed features
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        inputs=[
            if1,
            if2,
            if3,
            if4,
            if5,
            if6,
            if9,
        ],  # all feature types incl. with fixed values
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        inputs=[
            if1,
            if2,
            if3,
            if4,
            if5,
            if6,
            if9,
        ],  # all feature types incl. with fixed values + mutli-objective
        outputs=[of1, of2],
        constraints=[],
    ),
    Domain.from_lists(
        inputs=[if1, if2],  # only continuous features
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        inputs=[if1, if3, if5, if9],  # all feature types + mutli-objective
        outputs=[of1, of2],
        constraints=[],
    ),
    Domain.from_lists(
        inputs=[if1, if2, if8],
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        inputs=[if1, if2],  # only continuous features
        outputs=[of1, of2],
        constraints=[],
    )
    # Domain(
    #     inputs=[if1, if7], # unknown dummy feature
    #     outputs=[of1],
    #     constraints=[],
    # )
]

data = [
    pd.DataFrame.from_dict(
        {
            "if1": [3, 4, 5, 4.5],
            "if3": ["c1", "c2", "c3", "c1"],
            "if5": ["c1", "c2", "c3", "c1"],
            "if9": [1.0, 2.0, 1.0, 2.0],
            "of1": [10, 11, 12, 13],
            "valid_of1": [1, 0, 1, 0],
        }
    ),
    pd.DataFrame.from_dict(
        {
            "if1": [3, 4, 5, 4.5],
            "if2": [3, 3, 3, 3],
            "if3": ["c1", "c2", "c3", "c1"],
            "if4": ["c1", "c1", "c1", "c1"],
            "if5": ["c1", "c2", "c3", "c1"],
            "if6": ["c1", "c1", "c1", "c1"],
            "if9": [1.0, 2.0, 1.0, 2.0],
            "of1": [10, 11, 12, 13],
            "valid_of1": [1, 0, 1, 0],
        }
    ),
    pd.DataFrame.from_dict(
        {
            "if1": [3, 4, 5, 4.5],
            "if2": [3, 3, 3, 3],
            "if3": ["c1", "c2", "c3", "c1"],
            "if4": ["c1", "c1", "c1", "c1"],
            "if5": ["c1", "c2", "c3", "c1"],
            "if6": ["c1", "c1", "c1", "c1"],
            "if9": [1.0, 2.0, 1.0, 2.0],
            "of1": [10, 11, 12, 13],
            "of2": [100, 103, 105, 110],
            "valid_of1": [1, 0, 1, 0],
            "valid_of2": [0, 1, 1, 0],
        }
    ),
    pd.DataFrame.from_dict(
        {
            "if1": [3, 4, 5, 4.5],
            "if2": [3, 3, 3, 3],
            "of1": [10, 11, 12, 13],
            "valid_of1": [1, 0, 1, 0],
        }
    ),
    pd.DataFrame.from_dict(
        {
            "if1": [3, 4, 5, 4.5],
            "if3": ["c1", "c2", "c3", "c1"],
            "if5": ["c1", "c2", "c3", "c1"],
            "if9": [1.0, 2.0, 1.0, 2.0],
            "of1": [10, 11, 12, 13],
            "of2": [100, 103, 105, 110],
            "valid_of1": [1, 0, 1, 0],
            "valid_of2": [0, 1, 1, 0],
        }
    ),
]


@pytest.mark.parametrize("domain", list(domains))
def test_base_create(domain: Domain):
    with pytest.raises(ValueError, match="number sobol samples"):
        DummyStrategyDataModel(domain=domain, num_sobol_samples=5)

    with pytest.raises(ValueError, match="number raw samples"):
        DummyStrategyDataModel(domain=domain, num_raw_samples=5)


def test_base_invalid_descriptor_method():
    with pytest.raises(ValueError):
        DummyStrategyDataModel(
            domain=domains[0],
            surrogate_specs=[
                surrogate_data_models.SingleTaskGPSurrogate(
                    inputs=domains[0].inputs,
                    outputs=domains[0].outputs,
                    input_preprocessing_specs={"if5": CategoricalEncodingEnum.ONE_HOT},
                )
            ],  # type: ignore
            descriptor_method="FREE",  # type: ignore
            categorical_method="EXHAUSTIVE",  # type: ignore
        )


@pytest.mark.parametrize(
    # "domain, descriptor_encoding, categorical_encoding, categorical_method, expected",
    "domain, surrogate_specs, categorical_method, descriptor_method, expected",
    [
        (domains[0], None, "EXHAUSTIVE", "EXHAUSTIVE", {}),
        (
            domains[1],
            None,
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            {1: 3, 5: 1, 6: 2, 10: 1, 11: 0, 12: 0},
        ),
        (
            domains[1],
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[1].inputs,
                        outputs=domains[1].outputs,
                        input_preprocessing_specs={
                            "if5": CategoricalEncodingEnum.ONE_HOT,
                            "if6": CategoricalEncodingEnum.ONE_HOT,
                        },
                    )
                ]
            ),
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            {1: 3, 6: 1, 7: 0, 8: 0, 12: 1, 13: 0, 14: 0},
        ),
        (
            domains[1],
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[1].inputs,
                        outputs=domains[1].outputs,
                    )
                ]
            ),
            "FREE",
            "EXHAUSTIVE",
            {1: 3, 5: 1, 6: 2, 10: 1, 11: 0, 12: 0},
        ),
        (  #
            domains[1],
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[1].inputs,
                        outputs=domains[1].outputs,
                        input_preprocessing_specs={
                            "if5": CategoricalEncodingEnum.ONE_HOT,
                            "if6": CategoricalEncodingEnum.ONE_HOT,
                        },
                    )
                ]
            ),
            "FREE",
            "FREE",
            {1: 3, 6: 1, 7: 0, 8: 0, 12: 1, 13: 0, 14: 0},
        ),
        (domains[5], None, "EXHAUSTIVE", "EXHAUSTIVE", {1: 3.0}),
        (
            domains[5],
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[5].inputs,
                        outputs=domains[5].outputs,
                        input_preprocessing_specs={
                            "if8": CategoricalEncodingEnum.ONE_HOT,
                        },
                    )
                ]
            ),
            "FREE",
            "FREE",
            {1: 3.0, 2: 0},
        ),
        (domains[5], None, "FREE", "EXHAUSTIVE", {1: 3.0}),
        (domains[5], None, "FREE", "FREE", {1: 3.0, 2: 3.0}),
    ],
)
def test_base_get_fixed_features(
    domain, surrogate_specs, categorical_method, descriptor_method, expected
):
    data_model = DummyStrategyDataModel(
        domain=domain,
        surrogate_specs=surrogate_specs,
        categorical_method=categorical_method,
        descriptor_method=descriptor_method,
    )
    myStrategy = DummyStrategy(data_model=data_model)

    experiments = generate_experiments(domain, 100, tol=1.0)
    myStrategy.set_experiments(experiments)

    fixed_features = myStrategy.get_fixed_features()

    assert fixed_features == expected


@pytest.mark.parametrize(
    "domain, descriptor_method, categorical_method, discrete_method, surrogate_specs, expected",
    [
        (
            domains[0],
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            None,
            [
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 1.0, 3: 2.0, 1: 1},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 3.0, 3: 7.0, 1: 1},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 5.0, 3: 1.0, 1: 1},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 1.0, 3: 2.0, 1: 1},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 3.0, 3: 7.0, 1: 1},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 5.0, 3: 1.0, 1: 1},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 1.0, 3: 2.0, 1: 1},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 3.0, 3: 7.0, 1: 1},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 5.0, 3: 1.0, 1: 1},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 1.0, 3: 2.0, 1: 2},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 3.0, 3: 7.0, 1: 2},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 5.0, 3: 1.0, 1: 2},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 1.0, 3: 2.0, 1: 2},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 3.0, 3: 7.0, 1: 2},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 5.0, 3: 1.0, 1: 2},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 1.0, 3: 2.0, 1: 2},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 3.0, 3: 7.0, 1: 2},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 5.0, 3: 1.0, 1: 2},
            ],
        ),
        (
            domains[0],
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            "FREE",
            None,
            [
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 1.0, 3: 2.0},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 3.0, 3: 7.0},
                {4: 1.0, 5: 0.0, 6: 0.0, 2: 5.0, 3: 1.0},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 1.0, 3: 2.0},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 3.0, 3: 7.0},
                {4: 0.0, 5: 1.0, 6: 0.0, 2: 5.0, 3: 1.0},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 1.0, 3: 2.0},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 3.0, 3: 7.0},
                {4: 0.0, 5: 0.0, 6: 1.0, 2: 5.0, 3: 1.0},
            ],
        ),
        (
            domains[0],
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[0].inputs,
                        outputs=domains[0].outputs,
                        input_preprocessing_specs={
                            "if5": CategoricalEncodingEnum.ONE_HOT,
                        },
                    )
                ]
            ),
            [
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 1.0, 6: 0.0, 7: 0.0, 1: 1},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 0.0, 6: 1.0, 7: 0.0, 1: 1},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 0.0, 6: 0.0, 7: 1.0, 1: 1},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 1.0, 6: 0.0, 7: 0.0, 1: 1},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 0.0, 6: 1.0, 7: 0.0, 1: 1},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 0.0, 6: 0.0, 7: 1.0, 1: 1},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 1.0, 6: 0.0, 7: 0.0, 1: 1},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 0.0, 6: 1.0, 7: 0.0, 1: 1},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 0.0, 6: 0.0, 7: 1.0, 1: 1},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 1.0, 6: 0.0, 7: 0.0, 1: 2},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 0.0, 6: 1.0, 7: 0.0, 1: 2},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 0.0, 6: 0.0, 7: 1.0, 1: 2},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 1.0, 6: 0.0, 7: 0.0, 1: 2},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 0.0, 6: 1.0, 7: 0.0, 1: 2},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 0.0, 6: 0.0, 7: 1.0, 1: 2},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 1.0, 6: 0.0, 7: 0.0, 1: 2},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 0.0, 6: 1.0, 7: 0.0, 1: 2},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 0.0, 6: 0.0, 7: 1.0, 1: 2},
            ],
        ),
        (
            domains[0],
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            "FREE",
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[0].inputs,
                        outputs=domains[0].outputs,
                        input_preprocessing_specs={
                            "if5": CategoricalEncodingEnum.ONE_HOT,
                        },
                    )
                ]
            ),
            [
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 1.0, 6: 0.0, 7: 0.0},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 0.0, 6: 1.0, 7: 0.0},
                {2: 1.0, 3: 0.0, 4: 0.0, 5: 0.0, 6: 0.0, 7: 1.0},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 1.0, 6: 0.0, 7: 0.0},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 0.0, 6: 1.0, 7: 0.0},
                {2: 0.0, 3: 1.0, 4: 0.0, 5: 0.0, 6: 0.0, 7: 1.0},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 1.0, 6: 0.0, 7: 0.0},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 0.0, 6: 1.0, 7: 0.0},
                {2: 0.0, 3: 0.0, 4: 1.0, 5: 0.0, 6: 0.0, 7: 1.0},
            ],
        ),
        (
            domains[0],
            "EXHAUSTIVE",
            "FREE",
            "FREE",
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[0].inputs,
                        outputs=domains[0].outputs,
                    )
                ]
            ),
            [{2: 1.0, 3: 2.0}, {2: 3.0, 3: 7.0}, {2: 5.0, 3: 1.0}],
        ),
        (
            domains[0],
            "FREE",
            "FREE",
            "EXHAUSTIVE",
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[0].inputs,
                        outputs=domains[0].outputs,
                    )
                ]
            ),
            [{1: 1.0}, {1: 2.0}],
        ),
        (
            domains[0],
            "EXHAUSTIVE",
            "FREE",
            "EXHAUSTIVE",
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[0].inputs,
                        outputs=domains[0].outputs,
                    )
                ]
            ),
            [
                {2: 1.0, 3: 2.0, 1: 1.0},
                {2: 3.0, 3: 7.0, 1: 1.0},
                {2: 5.0, 3: 1.0, 1: 1.0},
                {2: 1.0, 3: 2.0, 1: 2.0},
                {2: 3.0, 3: 7.0, 1: 2.0},
                {2: 5.0, 3: 1.0, 1: 2.0},
            ],
        ),
        (
            domains[0],
            "FREE",
            "FREE",
            "FREE",
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=domains[0].inputs,
                        outputs=domains[0].outputs,
                        input_preprocessing_specs={
                            "if5": CategoricalEncodingEnum.ONE_HOT
                        },
                    )
                ]
            ),
            [{}],
        ),
        (
            domains[0],
            "FREE",
            "EXHAUSTIVE",
            "FREE",
            None,
            [
                {4: 1.0, 5: 0.0, 6: 0.0},
                {4: 0.0, 5: 1.0, 6: 0.0},
                {4: 0.0, 5: 0.0, 6: 1.0},
            ],
        ),
        (
            domains[3],
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            "EXHAUSTIVE",
            None,
            [
                {1: 3.0},
            ],
        ),
    ],
)
def test_base_get_categorical_combinations(
    domain,
    descriptor_method,
    categorical_method,
    discrete_method,
    surrogate_specs,
    expected,
):
    data_model = DummyStrategyDataModel(
        domain=domain,
        surrogate_specs=surrogate_specs,
        descriptor_method=descriptor_method,
        categorical_method=categorical_method,
        discrete_method=discrete_method,
    )
    myStrategy = DummyStrategy(data_model=data_model)
    c = unittest.TestCase()
    combo = myStrategy.get_categorical_combinations()
    c.assertCountEqual(combo, expected)


@pytest.mark.parametrize("domain", [(domains[0])])
def test_base_invalid_pair_encoding_method(domain):
    with pytest.raises(ValueError):
        DummyStrategyDataModel(
            domain=domain, categorical_encoding="ORDINAL", categorical_method="FREE"  # type: ignore
        )


@pytest.mark.parametrize(
    "domain, data",
    [
        (
            domains[0],
            generate_experiments(
                domains[0], row_count=5, tol=1.0, force_all_categories=True
            ),
        ),
        (
            domains[1],
            generate_experiments(
                domains[1], row_count=5, tol=1.0, force_all_categories=True
            ),
        ),
        (
            domains[2],
            generate_experiments(
                domains[2], row_count=5, tol=1.0, force_all_categories=True
            ),
        ),
        (
            domains[4],
            generate_experiments(
                domains[4], row_count=5, tol=1.0, force_all_categories=True
            ),
        ),
    ],
)
def test_base_fit(domain, data):
    data_model = DummyStrategyDataModel(domain=domain)
    myStrategy = DummyStrategy(data_model=data_model)
    myStrategy.set_experiments(data)
    myStrategy.fit()


# TODO: replace this with proper benchmark methods
@pytest.mark.parametrize(
    "domain, data, acquisition_function",
    [
        (
            domains[0],
            generate_experiments(
                domains[0], row_count=10, tol=1.0, force_all_categories=True
            ),
            specs.acquisition_functions.valid().obj(),
        ),
        (
            domains[1],
            generate_experiments(
                domains[1], row_count=10, tol=1.0, force_all_categories=True
            ),
            specs.acquisition_functions.valid().obj(),
        ),
        # TODO: this tests randomly fails (All attempts to fit the model have failed.)
        # (
        #     domains[2],
        #     generate_experiments(
        #         domains[2], row_count=10, tol=1.0, force_all_categories=True
        #     ),
        #     specs.acquisition_functions.valid().obj(),
        # ),
        # TODO: this tests randomly fails (All attempts to fit the model have failed.)
        # (
        #     domains[4],
        #     generate_experiments(
        #         domains[4], row_count=10, tol=1.0, force_all_categories=True
        #     ),
        #     specs.acquisition_functions.valid().obj(),
        # ),
    ],
)
def test_base_predict(domain, data, acquisition_function):
    data_model = DummyStrategyDataModel(
        domain=domain
    )  # , acquisition_function=acquisition_function
    # )
    myStrategy = DummyStrategy(data_model=data_model)
    myStrategy.tell(experiments=data)
    predictions = myStrategy.predict(data)
    assert len(predictions.columns.tolist()) == 3 * len(domain.get_feature_keys(Output))
    assert data.index[-1] == predictions.index[-1]


@pytest.mark.parametrize(
    "categorical_method, descriptor_method, discrete_method",
    list(
        itertools.product(
            [CategoricalMethodEnum.FREE, CategoricalMethodEnum.EXHAUSTIVE],
            [CategoricalMethodEnum.FREE, CategoricalMethodEnum.EXHAUSTIVE],
            [CategoricalMethodEnum.FREE, CategoricalMethodEnum.EXHAUSTIVE],
        )
    ),
)
def test_base_setup_ask_fixed_features(
    categorical_method, descriptor_method, discrete_method
):
    # test for fixed features list
    data_model = DummyStrategyDataModel(
        domain=domains[0],
        # acquisition_function=specs.acquisition_functions.valid().obj(),
        categorical_method=categorical_method,
        descriptor_method=descriptor_method,
        discrete_method=discrete_method,
        surrogate_specs=surrogate_data_models.BotorchSurrogates(
            surrogates=[
                surrogate_data_models.SingleTaskGPSurrogate(
                    inputs=domains[0].inputs,
                    outputs=domains[0].outputs,
                    # input_preprocessing_specs={"if5": CategoricalEncodingEnum.ONE_HOT},
                )
            ]
        ),
    )
    myStrategy = DummyStrategy(data_model=data_model)
    (
        bounds,
        ic_generator,
        ic_gen_kwargs,
        nchooseks,
        fixed_features,
        fixed_features_list,
    ) = myStrategy._setup_ask()
    if any(
        enc == CategoricalMethodEnum.EXHAUSTIVE
        for enc in [
            categorical_method,
            descriptor_method,
            discrete_method,
        ]
    ):
        assert fixed_features is None
        assert fixed_features_list is not None
    else:
        assert fixed_features == {}
        assert fixed_features_list is None

    data_model = DummyStrategyDataModel(
        domain=domains[3],
        # acquisition_function=specs.acquisition_functions.valid().obj(),
        categorical_method=categorical_method,
        descriptor_method=descriptor_method,
        discrete_method=discrete_method,
    )
    myStrategy = DummyStrategy(data_model=data_model)
    (
        bounds,
        ic_generator,
        ic_gen_kwargs,
        nchooseks,
        fixed_features,
        fixed_features_list,
    ) = myStrategy._setup_ask()
    assert fixed_features == {1: 3.0}
    assert fixed_features_list is None


def test_base_setup_ask():
    # test for no nchooseks
    benchmark = Hartmann()
    data_model = DummyStrategyDataModel(
        domain=benchmark.domain,
        # acquisition_function=specs.acquisition_functions.valid().obj(),
    )
    myStrategy = DummyStrategy(data_model=data_model)
    (
        bounds,
        ic_generator,
        ic_gen_kwargs,
        nchooseks,
        fixed_features,
        fixed_features_list,
    ) = myStrategy._setup_ask()
    assert torch.allclose(
        bounds,
        torch.tensor([[0 for _ in range(6)], [1 for _ in range(6)]]).to(**tkwargs),
    )
    assert ic_generator is None
    assert ic_gen_kwargs == {}
    assert nchooseks is None
    assert fixed_features == {}
    assert fixed_features_list is None
    # test for nchooseks
    benchmark = Hartmann(dim=6, allowed_k=3)
    data_model = DummyStrategyDataModel(
        domain=benchmark.domain,
        # acquisition_function=specs.acquisition_functions.valid().obj(),
    )
    myStrategy = DummyStrategy(data_model=data_model)
    (
        bounds,
        ic_generator,
        ic_gen_kwargs,
        nchooseks,
        fixed_features,
        fixed_features_list,
    ) = myStrategy._setup_ask()
    assert torch.allclose(
        bounds,
        torch.tensor([[0 for _ in range(6)], [1 for _ in range(6)]]).to(**tkwargs),
    )
    assert ic_generator == gen_batch_initial_conditions
    assert list(ic_gen_kwargs.keys()) == ["generator"]
    assert len(nchooseks) == len(get_nchoosek_constraints(domain=benchmark.domain))  # type: ignore
    assert fixed_features == {}
    assert fixed_features_list is None



from typing import List, Literal

import numpy as np
import pandas as pd
from pydantic import root_validator, validator

from bofire.data_models.constraints.constraint import Constraint, FeatureKeys


def narrow_gaussian(x, ell=1e-3):
    return np.exp(-0.5 * (x / ell) ** 2)


class NChooseKConstraint(Constraint):
    """NChooseK constraint that defines how many ingredients are allowed in a formulation.

    Attributes:
        features (List[str]): List of feature keys to which the constraint applies.
        min_count (int): Minimal number of non-zero/active feature values.
        max_count (int): Maximum number of non-zero/active feature values.
        none_also_valid (bool): In case that min_count > 0,
            this flag decides if zero active features are also allowed.
    """

    type: Literal["NChooseKConstraint"] = "NChooseKConstraint"
    features: FeatureKeys
    min_count: int
    max_count: int
    none_also_valid: bool

    @validator("features")
    def validate_features_unique(cls, features: List[str]):
        """Validates that provided feature keys are unique."""
        if len(features) != len(set(features)):
            raise ValueError("features must be unique")
        return features

    @root_validator(pre=False, skip_on_failure=True)
    def validate_counts(cls, values):
        """Validates if the minimum and maximum of allowed features are smaller than the overall number of features."""
        features = values["features"]
        min_count = values["min_count"]
        max_count = values["max_count"]

        if min_count > len(features):
            raise ValueError("min_count must be <= # of features")
        if max_count > len(features):
            raise ValueError("max_count must be <= # of features")
        if min_count > max_count:
            raise ValueError("min_values must be <= max_values")

        return values

    def __call__(self, experiments: pd.DataFrame) -> pd.Series:
        """Smooth relaxation of NChooseK constraint by countig the number of zeros in a candidate by a sum of
        narrow gaussians centered at zero.

        Args:
            experiments (pd.DataFrame): Data to evaluate the constraint on.

        Returns:
            pd.Series containing the constraint violation for each experiment (row in experiments argument).
        """

        def relu(x):
            return np.maximum(0, x)

        indices = np.array(
            [i for i, name in enumerate(experiments.columns) if name in self.features],
            dtype=np.int64,
        )
        experiments_tensor = np.array(experiments.to_numpy())

        max_count_violation = np.zeros(experiments_tensor.shape[0])
        min_count_violation = np.zeros(experiments_tensor.shape[0])

        if self.max_count != len(self.features):
            max_count_violation = relu(-1 * narrow_gaussian(x=experiments_tensor[..., indices]).sum(axis=-1) + (len(self.features) - self.max_count))  # type: ignore

        if self.min_count > 0:
            min_count_violation = relu(narrow_gaussian(x=experiments_tensor[..., indices]).sum(axis=-1) - (len(self.features) - self.min_count))  # type: ignore

        return pd.Series(max_count_violation + min_count_violation)

    def is_fulfilled(self, experiments: pd.DataFrame, tol: float = 1e-6) -> pd.Series:
        """Check if the concurrency constraint is fulfilled for all the rows of the provided dataframe.

        Args:
            experiments (pd.DataFrame): Dataframe to evaluate constraint on.
            tol (float,optional): tolerance parameter. A constraint is considered as not fulfilled
                if the violation is larger than tol. Defaults to 1e-6.

        Returns:
            bool: True if fulfilled else False.
        """

        cols = self.features
        sums = (np.abs(experiments[cols]) > tol).sum(axis=1)

        lower = sums >= self.min_count
        upper = sums <= self.max_count

        if not self.none_also_valid:
            # return lower.all() and upper.all()
            return pd.Series(np.logical_and(lower, upper), index=experiments.index)
        else:
            none = sums == 0
            return pd.Series(
                np.logical_or(none, np.logical_and(lower, upper)),
                index=experiments.index,
            )

    def __str__(self):
        """Generate string representation of the constraint.

        Returns:
            str: string representation of the constraint.
        """
        res = (
            "of the features "
            + ", ".join(self.features)
            + f" between {self.min_count} and {self.max_count} must be used"
        )
        if self.none_also_valid:
            res += " (none is also ok)"
        return res

    def jacobian(self, experiments: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("Jacobian not implemented for NChooseK constraints.")



import pytest

import bofire.data_models.strategies.api as data_models
import bofire.strategies.api as strategies
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Constraints, Domain, Inputs
from bofire.data_models.features.api import CategoricalInput, ContinuousInput

inputs = Inputs(
    features=[ContinuousInput(key=f"if{i}", bounds=(0, 1)) for i in range(1, 4)]
)
constraints = Constraints(
    constraints=[
        LinearInequalityConstraint(
            features=["if1", "if2", "if3"], coefficients=[1, 1, 1], rhs=1
        )
    ]
)


@pytest.mark.parametrize(
    "features, constraints, sampling_method, num_samples",
    [
        (inputs, constraints, sampling_method, num_samples)
        for sampling_method in ["SOBOL", "UNIFORM", "LHS"]
        for num_samples in [1, 2, 64, 128]
    ],
)
def test_rejection_sampler(features, constraints, sampling_method, num_samples):
    domain = Domain(
        inputs=features,
        constraints=constraints,
    )
    data_model = data_models.RejectionSampler(
        domain=domain, sampling_method=sampling_method
    )
    sampler = strategies.RejectionSampler(data_model=data_model)
    sampler.ask(num_samples)


def test_rejection_sampler_not_converged():
    domain = Domain(
        inputs=inputs,
        constraints=constraints,
    )
    data_model = data_models.RejectionSampler(
        domain=domain, num_base_samples=16, max_iters=2
    )
    sampler = strategies.RejectionSampler(data_model=data_model)
    with pytest.raises(ValueError):
        sampler.ask(128)


if1 = ContinuousInput(
    bounds=(0, 1),
    key="if1",
)
if2 = ContinuousInput(
    bounds=(0, 1),
    key="if2",
)
if3 = ContinuousInput(
    bounds=(0, 1),
    key="if3",
)
if4 = ContinuousInput(
    bounds=(0.1, 0.1),
    key="if4",
)
if5 = CategoricalInput(
    categories=["a", "b", "c"],
    key="if5",
)
if6 = CategoricalInput(
    categories=["a", "b", "c"],
    allowed=[False, True, False],
    key="if6",
)
If7 = ContinuousInput(bounds=(1, 1), key="If7")
c1 = LinearEqualityConstraint(
    features=["if1", "if2", "if3", "if4"], coefficients=[1.0, 1.0, 1.0, 1.0], rhs=1.0
)
c2 = LinearInequalityConstraint.from_greater_equal(
    features=["if1", "if2"], coefficients=[1.0, 1.0], rhs=0.2
)
c3 = LinearInequalityConstraint.from_greater_equal(
    features=["if1", "if2", "if4"], coefficients=[1.0, 1.0, 0.5], rhs=0.2
)
c4 = LinearEqualityConstraint(
    features=["if1", "if2", "if3", "if4", "If7"],
    coefficients=[1.0, 1.0, 1.0, 1.0, 1.0],
    rhs=2.0,
)
c5 = LinearEqualityConstraint(features=["if1", "if4"], coefficients=[1.0, 1.0], rhs=1.0)
c6 = NChooseKConstraint(
    features=["if1", "if2", "if3"],
    min_count=1,
    max_count=2,
    none_also_valid=False,
)

domains = [
    Domain.from_lists(inputs=[if1, if2, if3], constraints=[c2]),
    Domain.from_lists(inputs=[if1, if2, if3, if4], constraints=[c1, c2]),
    Domain.from_lists(inputs=[if1, if2, if3, if4], constraints=[c1, c2, c3]),
    Domain.from_lists(inputs=[if1, if2, if3, if5], constraints=[c2]),
    Domain.from_lists(inputs=[if1, if2, if3, if4, if5], constraints=[c1, c2]),
    Domain.from_lists(inputs=[if1, if2, if3, if4, if5], constraints=[c1, c2, c3]),
    Domain.from_lists(inputs=[if1, if2, if3, if6], constraints=[c2]),
    Domain.from_lists(inputs=[if1, if2, if3, if4, if6], constraints=[c1, c2]),
    Domain.from_lists(inputs=[if1, if2, if3, if4, if6], constraints=[c1, c2, c3]),
    Domain.from_lists(inputs=[if1, if2, if3, if4, if6, If7], constraints=[c1, c2, c3]),
    Domain.from_lists(
        inputs=[if1, if2, if3, if4, if6, If7], constraints=[c1, c2, c3, c4]
    ),
    Domain.from_lists(inputs=[if1, if2, if3, if4], constraints=[c5]),
    Domain.from_lists(
        inputs=[if1, if2, if3, if4, if6, If7],
        constraints=[c6, c2],
    ),
    Domain.from_lists(
        inputs=[if1, if2, if3, if4, if6, If7],
        constraints=[c6, c2, c1],
    ),
]


@pytest.mark.parametrize(
    "domain, candidate_count",
    [
        (domain, candidate_count)
        for domain in domains
        for candidate_count in range(1, 2)
    ],
)
def test_PolytopeSampler(domain, candidate_count):
    data_model = data_models.PolytopeSampler(domain=domain)
    sampler = strategies.PolytopeSampler(data_model=data_model)
    samples = sampler.ask(candidate_count)
    if len(domain.constraints.get(NChooseKConstraint)) == 0:
        assert len(samples) == candidate_count


def test_PolytopeSampler_all_fixed():
    domain = Domain.from_lists(inputs=[if1, if4], constraints=[c5])
    data_model = data_models.PolytopeSampler(domain=domain)
    sampler = strategies.PolytopeSampler(data_model=data_model)
    with pytest.warns(UserWarning):
        sampler.ask(2)


def test_PolytopeSampler_nchoosek():
    domain = Domain.from_lists(
        inputs=[if1, if2, if3, if4, if6, If7],
        constraints=[c6, c2],
    )
    data_model = data_models.PolytopeSampler(domain=domain)
    sampler = strategies.PolytopeSampler(data_model=data_model)
    samples = sampler.ask(5, return_all=True)
    assert len(samples) == 15
    samples = sampler.ask(50, return_all=False)
    assert len(samples) == 50



from typing import List

import pandas as pd
import pytest

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)

F = FEATURES = ["f" + str(i) for i in range(1, 11)]

C = list(range(1, 11))


def get_row(features, value: float = None, values: List[float] = None):
    if values is None:
        values = [value for _ in range(len(features))]
    return pd.DataFrame.from_dict([dict(zip(features, values))])


@pytest.mark.parametrize(
    "df, constraint, fulfilled",
    [
        (
            get_row(F[:4], 1),
            NonlinearEqualityConstraint(
                expression="f1 + f2 + f3 + f4 -4",
            ),
            True,
        ),
        (
            get_row(F[:4], 1),
            NonlinearEqualityConstraint(
                expression="f1 + f2 + f3 + f4 -3",
            ),
            False,
        ),
        (
            get_row(F[:4], 1),
            NonlinearInequalityConstraint(
                expression="f1 + f2 + f3 + f4 -5",
            ),
            True,
        ),
        (
            get_row(F[:4], 1),
            NonlinearInequalityConstraint(
                expression="f1 + f2 + f3 + f4 -2",
            ),
            False,
        ),
        (
            get_row(F[:4], 1),
            LinearEqualityConstraint(features=F[:4], coefficients=C[:4], rhs=10),
            True,
        ),
        (
            pd.concat([get_row(F[:4], 1), get_row(F[:4], 1)], ignore_index=True),
            LinearEqualityConstraint(features=F[:4], coefficients=C[:4], rhs=10),
            True,
        ),
        (
            pd.concat([get_row(F[:4], 1), get_row(F[:4], 2)]),
            LinearEqualityConstraint(features=F[:4], coefficients=C[:4], rhs=10),
            False,
        ),
        (
            get_row(F[:4], 2),
            LinearEqualityConstraint(features=F[:4], coefficients=C[:4], rhs=20),
            True,
        ),
        (
            get_row(F[:10], 1),
            LinearEqualityConstraint(features=F[:10], coefficients=C[:10], rhs=55.001),
            False,
        ),
        (
            get_row(F[:10], 1),
            LinearEqualityConstraint(features=F[:10], coefficients=C[:10], rhs=54.999),
            False,
        ),
        (
            get_row(F[:3], 1),
            LinearInequalityConstraint.from_greater_equal(
                features=F[:3], coefficients=C[:3], rhs=6
            ),
            True,
        ),
        (
            pd.concat([get_row(F[:3], 1), get_row(F[:3], 1), get_row(F[:3], 2)]),
            LinearInequalityConstraint.from_greater_equal(
                features=F[:3], coefficients=C[:3], rhs=6
            ),
            True,
        ),
        (
            pd.concat([get_row(F[:3], 1), get_row(F[:3], 0.5)]),
            LinearInequalityConstraint.from_greater_equal(
                features=F[:3], coefficients=C[:3], rhs=6
            ),
            False,
        ),
        (
            get_row(F[:3], 1),
            LinearInequalityConstraint.from_greater_equal(
                features=F[:3], coefficients=C[:3], rhs=2
            ),
            True,
        ),
        (
            get_row(F[:3], 1),
            LinearInequalityConstraint.from_greater_equal(
                features=F[:3], coefficients=C[:3], rhs=6.001
            ),
            False,
        ),
        (
            get_row(F[:3], values=[1, 1, 1]),
            NChooseKConstraint(
                features=F[:3],
                min_count=0,
                max_count=3,
                none_also_valid=True,
            ),
            True,
        ),
        (
            pd.concat(
                [get_row(F[:3], values=[1, 1, 1]), get_row(F[:3], values=[1, 1, 1])]
            ),
            NChooseKConstraint(
                features=F[:3],
                min_count=0,
                max_count=3,
                none_also_valid=True,
            ),
            True,
        ),
        (
            get_row(F[:3], values=[0, 2, 3]),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=3,
                none_also_valid=False,
            ),
            True,
        ),
        (
            get_row(F[:3], values=[1, 2, 3]),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=2,
                none_also_valid=False,
            ),
            False,
        ),
        (
            get_row(F[:3], values=[0, 0, 3]),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=3,
                none_also_valid=False,
            ),
            False,
        ),
        (
            get_row(F[:3], values=[0, 0, 0]),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=3,
                none_also_valid=False,
            ),
            False,
        ),
        (
            get_row(F[:3], values=[0, 0, 0]),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=3,
                none_also_valid=True,
            ),
            True,
        ),
        (
            pd.concat(
                [get_row(F[:3], values=[0, 2, 3]), get_row(F[:3], values=[0, 0, 0])]
            ),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=2,
                none_also_valid=False,
            ),
            False,
        ),
        (
            pd.concat(
                [get_row(F[:3], values=[0, 2, 3]), get_row(F[:3], values=[0, 0, 0])]
            ),
            NChooseKConstraint(
                features=F[:3],
                min_count=2,
                max_count=2,
                none_also_valid=True,
            ),
            True,
        ),
    ],
)
def test_fulfillment(df, constraint, fulfilled):
    assert constraint.is_fulfilled(df).all() == fulfilled



import numpy as np
import pandas as pd

from bofire.data_models.domain.api import Domain
from bofire.data_models.strategies.api import RejectionSampler as DataModel
from bofire.strategies.samplers.sampler import SamplerStrategy


class RejectionSampler(SamplerStrategy):
    """Sampler that generates samples via rejection sampling.

    Attributes:
        domain (Domain): Domain defining the constrained input space
        sampling_method (SamplingMethodEnum, optional): Method to generate the base samples. Defaults to UNIFORM.
        num_base_samples (int, optional): Number of base samples to sample in each iteration. Defaults to 1000.
        max_iters (int, optinal): Number of iterations. Defaults to 1000.
    """

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.sampling_method = data_model.sampling_method
        self.num_base_samples = data_model.num_base_samples
        self.max_iters = data_model.max_iters

    def _ask(self, n: int) -> pd.DataFrame:
        if len(self.domain.constraints) == 0:
            return self.domain.inputs.sample(n, self.sampling_method)
        n_iters = 0
        n_found = 0
        valid_samples = []
        while n_found < n:
            if n_iters > self.max_iters:
                raise ValueError("Maximum iterations exceeded in rejection sampling.")
            samples = self.domain.inputs.sample(
                self.num_base_samples, method=self.sampling_method
            )
            valid = self.domain.constraints.is_fulfilled(samples)
            n_found += np.sum(valid)
            valid_samples.append(samples[valid])
            n_iters += 1
        return pd.concat(valid_samples, ignore_index=True).iloc[:n]

    def duplicate(self, domain: Domain) -> SamplerStrategy:
        data_model = DataModel(
            domain=domain,
            sampling_method=self.sampling_method,
            num_base_samples=self.num_base_samples,
            max_iters=self.max_iters,
        )
        return self.__class__(data_model=data_model)



import warnings
from abc import abstractmethod
from typing import List, Literal, Optional, Union

import pandas as pd
from pydantic import Field, root_validator, validator
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.api import Domain, Inputs, Outputs
from bofire.data_models.enum import RegressionMetricsEnum, UQRegressionMetricsEnum
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.objectives.api import MaximizeObjective, MinimizeObjective

metrics2objectives = {
    RegressionMetricsEnum.MAE: MinimizeObjective,
    RegressionMetricsEnum.MAPE: MinimizeObjective,
    RegressionMetricsEnum.MSD: MinimizeObjective,
    RegressionMetricsEnum.R2: MaximizeObjective,
    RegressionMetricsEnum.PEARSON: MaximizeObjective,
    RegressionMetricsEnum.SPEARMAN: MaximizeObjective,
    RegressionMetricsEnum.FISHER: MaximizeObjective,
    UQRegressionMetricsEnum.PEARSON_UQ: MaximizeObjective,
    UQRegressionMetricsEnum.SPEARMAN_UQ: MaximizeObjective,
    UQRegressionMetricsEnum.KENDALL_UQ: MaximizeObjective,
    UQRegressionMetricsEnum.MAXIMUMCALIBRATION: MinimizeObjective,
    UQRegressionMetricsEnum.MISCALIBRATIONAREA: MinimizeObjective,
    UQRegressionMetricsEnum.ABSOLUTEMISCALIBRATIONAREA: MinimizeObjective,
}


class Aggregation(BaseModel):
    type: str
    features: Annotated[List[str], Field(min_items=2)]
    keep_features: bool = False


class SumAggregation(Aggregation):
    type: Literal["SumAggregation"] = "SumAggregation"


class MeanAggregation(Aggregation):
    type: Literal["MeanAggregation"] = "MeanAggregation"


AnyAggregation = Union[SumAggregation, MeanAggregation]


class Hyperconfig(BaseModel):
    type: str
    hyperstrategy: Literal["RandomStrategy", "FactorialStrategy", "SoboStrategy"]
    inputs: Inputs
    n_iterations: Optional[Annotated[int, Field(ge=1)]] = None
    target_metric: RegressionMetricsEnum = RegressionMetricsEnum.MAE

    @validator("n_iterations", always=True)
    def validate_n_iterations(cls, v, values):
        if v is None:
            if values["hyperstrategy"] == "FactorialStrategy":
                return v
            return len(values["inputs"]) + 10
        else:
            if values["hyperstrategy"] == "FactorialStrategy":
                raise ValueError(
                    "It is not allowed to scpecify the number of its for FactorialStrategy"
                )
            if v < len(values["inputs"]) + 2:
                raise ValueError(
                    "At least number of hyperparams plus 2 iterations has to be specified"
                )
        return v

    @property
    def domain(self) -> Domain:
        return Domain(
            inputs=self.inputs,
            outputs=Outputs(
                features=[
                    ContinuousOutput(
                        key=self.target_metric.name,
                        objective=metrics2objectives[self.target_metric](),
                    )
                ]
            ),
        )

    @staticmethod
    @abstractmethod
    def _update_hyperparameters(surrogate_data, hyperparameters: pd.Series):
        pass


class TrainableSurrogate(BaseModel):
    hyperconfig: Optional[Hyperconfig] = None
    aggregations: Optional[Annotated[List[AnyAggregation], Field(min_items=1)]] = None

    @root_validator
    def validate_aggregations(cls, values):
        if values["aggregations"] is None:
            return values
        for agg in values["aggregations"]:
            for key in agg.features:
                if key not in values["inputs"].get_keys():
                    raise ValueError(
                        f"Unkown feature key {key} provided in aggregations."
                    )
                feat = values["inputs"].get_by_key(key)
                if not isinstance(feat, ContinuousInput):
                    raise ValueError(
                        f"Feature with key {key} is not of type ContinuousInput"
                    )
        warnings.warn("Aggregations currently only implemented in the data models.")
        return values

    def update_hyperparameters(self, hyperparameters: pd.Series):
        if self.hyperconfig is not None:
            self.hyperconfig.domain.validate_candidates(
                pd.DataFrame(hyperparameters).T,
                only_inputs=True,
                raise_validation_error=True,
            )
            self.hyperconfig._update_hyperparameters(
                self, hyperparameters=hyperparameters
            )
        else:
            raise ValueError("No hyperconfig available.")



import pytest

import bofire.strategies.api as strategies
from bofire.data_models.domain.api import Domain, Inputs
from bofire.data_models.features.api import CategoricalInput, DiscreteInput
from bofire.data_models.strategies.api import FactorialStrategy


def test_FactorialStrategy_ask():
    strategy_data = FactorialStrategy(
        domain=Domain(
            inputs=Inputs(
                features=[
                    CategoricalInput(key="alpha", categories=["a", "b", "c"]),
                    DiscreteInput(key="beta", values=[1.0, 2, 3.0, 4.0]),
                ]
            )
        )
    )
    strategy = strategies.map(strategy_data)
    candidates = strategy.ask(None)
    assert len(candidates) == 12


def test_FactorialStrategy_ask_invalid():
    strategy_data = FactorialStrategy(
        domain=Domain(
            inputs=Inputs(
                features=[
                    CategoricalInput(key="alpha", categories=["a", "b", "c"]),
                    DiscreteInput(key="beta", values=[1.0, 2, 3.0, 4.0]),
                ]
            )
        )
    )
    strategy = strategies.map(strategy_data)
    with pytest.raises(
        ValueError,
        match="FactorialStrategy will ignore the specified value of candidate_count. "
        "The strategy automatically determines how many candidates to "
        "propose.",
    ):
        strategy.ask(5)



import base64
import io
import warnings
from typing import Optional

import torch
from botorch.models.deterministic import DeterministicModel

from bofire.data_models.surrogates.api import EmpiricalSurrogate as DataModel
from bofire.surrogates.botorch import BotorchSurrogate


class EmpiricalSurrogate(BotorchSurrogate):
    """All necessary functions has to be implemented in the model which can then be loaded
    from cloud pickle.

    Attributes:
        model (DeterministicModel): Botorch model instance.
    """

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    model: Optional[DeterministicModel] = None

    def _dumps(self) -> str:
        """Dumps the actual model to a string via pickle as this is not directly json serializable."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            import bofire.surrogates.cloudpickle_module as cloudpickle_module

            if len(w) == 1:
                raise ModuleNotFoundError("Cloudpickle is not available.")

        buffer = io.BytesIO()
        torch.save(self.model, buffer, pickle_module=cloudpickle_module)  # type: ignore
        return base64.b64encode(buffer.getvalue()).decode()
        # return codecs.encode(pickle.dumps(self.model), "base64").decode()

    def loads(self, data: str):
        """Loads the actual model from a base64 encoded pickle bytes object and writes it to the `model` attribute."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            import bofire.surrogates.cloudpickle_module as cloudpickle_module

            if len(w) == 1:
                raise ModuleNotFoundError("Cloudpickle is not available.")

        buffer = io.BytesIO(base64.b64decode(data.encode()))
        self.model = torch.load(buffer, pickle_module=cloudpickle_module)  # type: ignore



from typing import Union

from bofire.data_models.molfeatures.molfeatures import (  # BagOfCharacters
    Fingerprints,
    FingerprintsFragments,
    Fragments,
    MolFeatures,
    MordredDescriptors,
)

AbstractMolFeatures = MolFeatures

AnyMolFeatures = Union[
    Fingerprints,
    Fragments,
    FingerprintsFragments,
    # BagOfCharacters,
    MordredDescriptors,
]



from typing import Annotated, Optional, Type

from pydantic import Field, PositiveInt, root_validator, validator

from bofire.data_models.constraints.api import (
    Constraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.domain.api import Domain, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum, CategoricalMethodEnum
from bofire.data_models.features.api import CategoricalDescriptorInput, CategoricalInput
from bofire.data_models.outlier_detection.api import OutlierDetections
from bofire.data_models.strategies.predictives.predictive import PredictiveStrategy
from bofire.data_models.surrogates.api import (
    BotorchSurrogates,
    MixedSingleTaskGPSurrogate,
    SingleTaskGPSurrogate,
)


def is_power_of_two(n):
    return (n != 0) and (n & (n - 1) == 0)


class BotorchStrategy(PredictiveStrategy):
    num_sobol_samples: PositiveInt = 512
    num_restarts: PositiveInt = 8
    num_raw_samples: PositiveInt = 1024
    descriptor_method: CategoricalMethodEnum = CategoricalMethodEnum.EXHAUSTIVE
    categorical_method: CategoricalMethodEnum = CategoricalMethodEnum.EXHAUSTIVE
    discrete_method: CategoricalMethodEnum = CategoricalMethodEnum.EXHAUSTIVE
    surrogate_specs: Optional[BotorchSurrogates] = None
    # outlier detection params
    outlier_detection_specs: Optional[OutlierDetections] = None
    min_experiments_before_outlier_check: PositiveInt = 1
    frequency_check: PositiveInt = 1
    # hyperopt params
    frequency_hyperopt: Annotated[int, Field(ge=0)] = 0  # 0 indicates no hyperopt
    folds: int = 5

    @classmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        """Method to check if a specific constraint type is implemented for the strategy

        Args:
            my_type (Type[Constraint]): Constraint class

        Returns:
            bool: True if the constraint type is valid for the strategy chosen, False otherwise
        """
        if my_type in [NonlinearInequalityConstraint, NonlinearEqualityConstraint]:
            return False
        return True

    @validator("num_sobol_samples")
    def validate_num_sobol_samples(cls, v):
        if is_power_of_two(v) is False:
            raise ValueError(
                "number sobol samples have to be of the power of 2 to increase performance"
            )
        return v

    @validator("num_raw_samples")
    def validate_num_raw_samples(cls, v):
        if is_power_of_two(v) is False:
            raise ValueError(
                "number raw samples have to be of the power of 2 to increase performance"
            )
        return v

    @root_validator(pre=False, skip_on_failure=True)
    def update_surrogate_specs_for_domain(cls, values):
        """Ensures that a prediction model is specified for each output feature"""
        values["surrogate_specs"] = BotorchStrategy._generate_surrogate_specs(
            values["domain"],
            values["surrogate_specs"],
        )
        # we also have to checke here that the categorical method is compatible with the chosen models
        if values["categorical_method"] == CategoricalMethodEnum.FREE:
            for m in values["surrogate_specs"].surrogates:
                if isinstance(m, MixedSingleTaskGPSurrogate):
                    raise ValueError(
                        "Categorical method FREE not compatible with a a MixedSingleTaskGPModel."
                    )
        #  we also check that if a categorical with descriptor method is used as one hot encoded the same method is
        # used for the descriptor as for the categoricals
        for m in values["surrogate_specs"].surrogates:
            keys = m.inputs.get_keys(CategoricalDescriptorInput)
            for k in keys:
                if m.input_preprocessing_specs[k] == CategoricalEncodingEnum.ONE_HOT:
                    if values["categorical_method"] != values["descriptor_method"]:
                        raise ValueError(
                            "One-hot encoded CategoricalDescriptorInput features has to be treated with the same method as categoricals."
                        )
        return values

    @root_validator(pre=False, skip_on_failure=True)
    def validate_outlier_detection_specs_for_domain(cls, values):
        """Ensures that a outlier_detection model is specified for each output feature"""
        if values["outlier_detection_specs"] is not None:
            values["outlier_detection_specs"]._check_compability(
                inputs=values["domain"].inputs, outputs=values["domain"].outputs
            )
        return values

    @staticmethod
    def _generate_surrogate_specs(
        domain: Domain,
        surrogate_specs: Optional[BotorchSurrogates] = None,
    ) -> BotorchSurrogates:
        """Method to generate model specifications when no model specs are passed
        As default specification, a 5/2 matern kernel with automated relevance detection and normalization of the input features is used.
        Args:
            domain (Domain): The domain defining the problem to be optimized with the strategy
            surrogate_specs (List[ModelSpec], optional): List of model specification classes specifying the models to be used in the strategy. Defaults to None.
        Raises:
            KeyError: if there is a model spec for an unknown output feature
            KeyError: if a model spec has an unknown input feature
        Returns:
            List[ModelSpec]: List of model specification classes
        """
        existing_keys = (
            surrogate_specs.outputs.get_keys() if surrogate_specs is not None else []
        )
        non_exisiting_keys = list(set(domain.outputs.get_keys()) - set(existing_keys))
        _surrogate_specs = (
            surrogate_specs.surrogates if surrogate_specs is not None else []
        )
        for output_feature in non_exisiting_keys:
            if len(domain.inputs.get(CategoricalInput, exact=True)):
                _surrogate_specs.append(
                    MixedSingleTaskGPSurrogate(
                        inputs=domain.inputs,
                        outputs=Outputs(features=[domain.outputs.get_by_key(output_feature)]),  # type: ignore
                    )
                )
            else:
                _surrogate_specs.append(
                    SingleTaskGPSurrogate(
                        inputs=domain.inputs,
                        outputs=Outputs(features=[domain.outputs.get_by_key(output_feature)]),  # type: ignore
                    )
                )
        surrogate_specs = BotorchSurrogates(surrogates=_surrogate_specs)  # type: ignore
        surrogate_specs._check_compability(inputs=domain.inputs, outputs=domain.outputs)
        return surrogate_specs
